#The background is set with 40 plus the number of the color, and the foreground with 30

#These are the sequences need to get colored ouput
import logging

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

#The background is set with 40 plus the number of the color, and the foreground with 30


RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"

def _formatter_message(message: str, use_color = True):
    if use_color:
        message = message.replace("$RESET", RESET_SEQ).replace("$BOLD", BOLD_SEQ)
    else:
        message = message.replace("$RESET", "").replace("$BOLD", "")
    return message

COLORS = {
    'WARNING': YELLOW,
    'INFO': WHITE,
    'DEBUG': BLUE,
    'CRITICAL': YELLOW,
    'ERROR': RED
}

class ColoredFormatter(logging.Formatter):
    def __init__(self, msg, use_color = True):
        logging.Formatter.__init__(self, msg)
        self.use_color = use_color

    def format(self, record):
        levelname = record.levelname
        if self.use_color and levelname in COLORS:
            levelname_color = COLOR_SEQ % (30 + COLORS[levelname]) + levelname + RESET_SEQ
            record.levelname = levelname_color
        return logging.Formatter.format(self, record)


# Custom logger class with multiple destinations
class ColoredLogger(logging.Logger):
    FORMAT = "\n[%(levelname)-18s]\n$BOLD%(message)-s$RESET"
    COLOR_FORMAT = _formatter_message(FORMAT, True)
    def __init__(self, name):
        logging.Logger.__init__(self, name, logging.DEBUG)                

        color_formatter = ColoredFormatter(self.COLOR_FORMAT)

        console = logging.StreamHandler()
        console.setFormatter(color_formatter)

        self.addHandler(console)
        return

logging.setLoggerClass(ColoredLogger)

def get_logger(name: str):
    return logging.getLogger(name)



import unittest

from harmony_model_checker.harmony.DumpASTVisitor import DumpASTVisitor
from harmony_model_checker.h2py.h2py import h2py
from harmony_model_checker.compile import parse_string
import ast as past


class TestH2PyProgram(unittest.TestCase):

    def assert_h2py(self, harmony_code: str, python_code: str):
        dump_ast = DumpASTVisitor()
        harmony_ast = parse_string(harmony_code)

        try:
            python_ast = h2py(harmony_ast)

        except Exception as error:
            self.fail(f"""
Error while calling H2PyASTVisitor: {error}

Dump Harmony AST: {dump_ast(harmony_ast)}
""")

        try:
            unparsed = past.unparse(python_ast)

        except Exception as error:
            self.fail(f"""
Error while unparsing Python AST: {error}

Dump Harmony AST: {dump_ast(harmony_ast)}

Dumped Parsed Python AST: {past.dump(python_ast, indent=2)}

Expected Python AST: {past.dump(past.parse(python_code), indent=2)}
""")

        self.assertEqual(python_code.strip(), unparsed.strip(), f"""
Dump Harmony AST: {dump_ast(harmony_ast)}

Dumped Parsed Python AST: {past.dump(python_ast, indent=2)}

Expected Python AST: {past.dump(past.parse(python_code), indent=2)}
""")

    def assert_h2py_files(self, harmony_path: str, python_path: str):
        with open(harmony_path) as harmony_file:
            with open(python_path) as python_file:
                self.assert_h2py(harmony_file.read(), python_file.read())

    def test_print(self):
        self.assert_h2py_files(
            'tests/resources/h2py/print.hny',
            'tests/resources/h2py/print.py'
        )

    def test_func_1(self):
        self.assert_h2py_files(
            'tests/resources/h2py/func_1.hny',
            'tests/resources/h2py/func_1.py',
        )

    def test_binops(self):
        self.assert_h2py_files(
            'tests/resources/h2py/binops.hny',
            'tests/resources/h2py/binops.py',
        )

    def test_dict_assign(self):
        self.assert_h2py_files(
            'tests/resources/h2py/dict_assign.hny',
            'tests/resources/h2py/dict_assign.py',
        )

    def test_imports(self):
        self.assert_h2py_files(
            'tests/resources/h2py/imports.hny',
            'tests/resources/h2py/imports.py',
        )

    def test_local_assign_assert(self):
        self.assert_h2py_files(
            'tests/resources/h2py/local_assign_assert.hny',
            'tests/resources/h2py/local_assign_assert.py',
        )

    def test_local_assign(self):
        self.assert_h2py_files(
            'tests/resources/h2py/local_assign.hny',
            'tests/resources/h2py/local_assign.py',
        )

    def test_ptr_assign_1(self):
        self.assert_h2py_files(
            'tests/resources/h2py/ptr_assign_1.hny',
            'tests/resources/h2py/ptr_assign_1.py',
        )

    def test_ptr_assign_2(self):
        self.assert_h2py_files(
            'tests/resources/h2py/ptr_assign_2.hny',
            'tests/resources/h2py/ptr_assign_2.py',
        )

    def test_tuple_assign(self):
        self.assert_h2py_files(
            'tests/resources/h2py/tuple_assign.hny',
            'tests/resources/h2py/tuple_assign.py',
        )

    def test_h2py_name_conflict(self):
        self.assert_h2py_files(
            'tests/resources/h2py/h2py_name_conflict.hny',
            'tests/resources/h2py/h2py_name_conflict.py',
        )

    @unittest.skip('''
    h2py_name_conflict_2.py is not semantically equivalent to 
    h2py_name_conflict_2.hny because, in renaming 'H' to '_H', this conflicts 
    with the existing _H name.
    ''')
    def test_h2py_name_conflict_2(self):
        self.assert_h2py_files(
            'tests/resources/h2py/h2py_name_conflict_2.hny',
            'tests/resources/h2py/h2py_name_conflict_2.py',
        )

    def test_addr_1(self):
        self.assert_h2py_files(
            'tests/resources/h2py/addr_1.hny',
            'tests/resources/h2py/addr_1.py',
        )

    def test_if_1(self):
        self.assert_h2py_files(
            'tests/resources/h2py/if_1.hny',
            'tests/resources/h2py/if_1.py',
        )

    def test_if_else(self):
        self.assert_h2py_files(
            'tests/resources/h2py/if_else.hny',
            'tests/resources/h2py/if_else.py',
        )

    def test_if_elif_else(self):
        self.assert_h2py_files(
            'tests/resources/h2py/if_elif_else.hny',
            'tests/resources/h2py/if_elif_else.py',
        )

    def test_var_1(self):
        self.assert_h2py_files(
            'tests/resources/h2py/var_1.hny',
            'tests/resources/h2py/var_1.py',
        )

    def test_while_1(self):
        self.assert_h2py_files(
            'tests/resources/h2py/while_1.hny',
            'tests/resources/h2py/while_1.py',
        )

    def test_var_2(self):
        self.assert_h2py_files(
            'tests/resources/h2py/var_2.hny',
            'tests/resources/h2py/var_2.py',
        )
    
    def test_choose_1(self):
        self.assert_h2py_files(
            'tests/resources/h2py/choose_1.hny',
            'tests/resources/h2py/choose_1.py',
        )

    def test_ptr_assign_3(self):
        self.assert_h2py_files(
            'tests/resources/h2py/ptr_assign_3.hny',
            'tests/resources/h2py/ptr_assign_3.py',
        )

    def test_ptr_assign_4(self):
        self.assert_h2py_files(
            'tests/resources/h2py/ptr_assign_4.hny',
            'tests/resources/h2py/ptr_assign_4.py',
        )

    def test_addr_2(self):
        self.assert_h2py_files(
            'tests/resources/h2py/addr_2.hny',
            'tests/resources/h2py/addr_2.py',
        )

    def test_ptr_apply_1(self):
        self.assert_h2py_files(
            'tests/resources/h2py/ptr_apply_1.hny',
            'tests/resources/h2py/ptr_apply_1.py',
        )

    def test_addr_3(self):
        self.assert_h2py_files(
            'tests/resources/h2py/addr_3.hny',
            'tests/resources/h2py/addr_3.py',
        )

    def test_apply_1(self):
        self.assert_h2py_files(
            'tests/resources/h2py/apply_1.hny',
            'tests/resources/h2py/apply_1.py',
        )


if __name__ == '__main__':
    unittest.main()



import json
import sys

contexts = []

def verbose_kv(js):
    return (verbose_string(js["key"]), verbose_string(js["value"]))

def verbose_idx(js):
    return "[" + verbose_string(js) + "]"

def verbose_tuple(v):
    if v == []:
        return "()"
    lst = [ verbose_string(val) for val in v ]
    return "(" + ", ".join(lst) + ")"

def verbose_string(js):
    global contexts 

    type = js["type"]
    if type == "address":
        if "func" not in js:
            return "None"
        result = "?"
        func = js["func"]
        args = js["args"]
        if func["type"] == "pc":
            if int(func["value"]) == -1:
                result += args[0]["value"]
                args = args[1:]
            elif int(func["value"]) == -2:
                result += "@" + args[0]["value"]
                args = args[1:]
            else:
                result += verbose_string(func)
        else:
            result += verbose_string(func)
        return result + "".join([ verbose_idx(kv) for kv in args ])
    v = js["value"]
    if type == "bool":
        return v
    if type == "int":
        return str(v) if isinstance(v, int) else v
    if type == "atom":
        return json.dumps(v, ensure_ascii=False)
    if type == "set":
        if v == []:
            return "{}"
        lst = [ verbose_string(val) for val in v ]
        return "{ " + ", ".join(lst) + " }"
    if type == "list":
        if v == []:
            return "[]"
        lst = [ verbose_string(val) for val in v ]
        return "[ " + ", ".join(lst) + " ]"
    if type == "dict":
        if v == []:
            return "{:}"
        lst = [ verbose_kv(kv) for kv in v ]
        keys = [ k for k,v in lst ]
        if keys == [str(i) for i in range(len(v))]:
            return "[ " + ", ".join([v for k,v in lst]) + " ]" 
        else:
            return "{ " + ", ".join([k + ": " + v for k,v in lst]) + " }" 
    if type == "pc":
        return "PC(%s)"%v
    if type == "context":
        if v in contexts:
            return "C" + str(contexts.index(v) + 1)
        else:
            contexts += [v]
            return "C" + str(len(contexts))

def verbose_print_vars(f, d):
    print("{", end="", file=f)
    first = True
    for k, v in d.items():
        if first:
            first = False
        else:
            print(",", end="", file=f)
        print(" %s: %s"%(k, verbose_string(v)), end="", file=f)
    print(" }", file=f)

def print_context(f, ctx):
    for k, v in ctx.items():
        if k == "vars":
            print("    * vars: ", end="", file=f)
            verbose_print_vars(f, v)
        elif "type" in v:
            print("    * %s: %s"%(k, verbose_string(v)), file=f)
        else:
            print("    * %s: %s"%(k, str(v)), file=f)

def verbose_print_trace(f, trace):
    for i, call in enumerate(trace):
        if i != 0:
            print(" --> ", end="", file=f)
        print("%s"%call["method"], end="", file=f)
    print(file=f)

def get_mode(ctx):
    mode = ctx["mode"]
    if mode == "terminated":
        return mode
    if "atomic" in ctx and int(ctx["atomic"]) > 0:
        mode += " atomic"
    if "readonly" in ctx and int(ctx["readonly"]) > 0:
        mode += " readonly"
    if "interruptlevel" in ctx and int(ctx["interruptlevel"]) > 0:
        mode += " interrupts-disabled"
    return mode

class Summarize:
    def __init__(self):
        self.tid = None
        self.name = None
        self.start = 0
        self.lastmis = {}
        self.shared = {}
        self.step = 0
        self.stack = []
        self.contexts = []
        self.module = None
        self.stmt = None
        self.expr = None

    def dump_contexts(self, f, exclude):
        for ctx in self.contexts:
            if ctx["tid"] != exclude:
                mode = get_mode(ctx)
                if int(ctx["tid"]) == 0 and mode == "terminated":
                    continue
                print("    * T%s: (%s) "%(ctx["tid"], mode), end="", file=f)
                verbose_print_trace(f, ctx["trace"])
                if "next" in ctx:
                    self.about(ctx, "        * ", f)

    def about(self, ctx, prefix, f):
        nxt = ctx["next"]
        if nxt["type"] == "Frame":
            print("%sabout to " % prefix, end="", file=f)
            self.aboutFrame(nxt, f)
        elif nxt["type"] == "Load":
            print("%sabout to " % prefix, end="", file=f)
            self.aboutLoad(ctx["pc"], nxt, f)
        elif nxt["type"] == "Store":
            print("%sabout to " % prefix, end="", file=f)
            self.aboutStore(ctx["pc"], nxt, f)
        elif nxt["type"] == "Print":
            print("%sabout to " % prefix, end="", file=f)
            self.aboutPrint(ctx["pc"], nxt, f)
        elif nxt["type"] == "AtomicInc":
            print("%sabout to " % prefix, end="", file=f)
            self.aboutAtomicInc(ctx["pc"], f)
        elif nxt["type"] == "Assert":
            print("%sabout to " % prefix, end="", file=f)
            self.aboutAssert(ctx["pc"], f)
        # else:
        #     print(nxt, file=f)

    def aboutFrame(self, nxt, f):
        print("run method %s with argument %s"%(nxt["name"], verbose_string(nxt["value"])), file=f)

    def aboutLoad(self, pc, nxt, f):
        print("load variable %s in "%nxt["var"], end="", file=f)
        loc = self.locations[int(pc)]
        self.print_loc_basic(loc, f)
        print(file=f)

    def aboutStore(self, pc, nxt, f):
        print("store %s into %s in "%(verbose_string(nxt["value"]), nxt["var"]), end="", file=f)
        loc = self.locations[int(pc)]
        self.print_loc_basic(loc, f)
        print(file=f)

    def aboutPrint(self, pc, nxt, f):
        print("print %s in "%verbose_string(nxt["value"]), end="", file=f)
        loc = self.locations[int(pc)]
        self.print_loc_basic(loc, f)
        print(file=f)

    def aboutAtomicInc(self, pc, f):
        loc = self.locations[int(pc)]
        module = self.hvm["modules"][loc["module"]]
        print("execute atomic section in ", end="", file=f)
        loc = self.locations[int(pc)]
        self.print_loc_basic(loc, f)
        print(file=f)

    def aboutAssert(self, pc, f):
        loc = self.locations[int(pc)]
        module = self.hvm["modules"][loc["module"]]
        print("fail assertion %s:%s: %s"%(loc["module"], loc["line"], module["lines"][int(loc["line"]) - 1]), file=f)

    def print_loc_basic(self, loc, f):
        if loc["module"] == "__main__":
            print("line %d" % loc["line"], end="", file=f)
        else:
            print("line %s/%d" % (loc["module"], loc["line"]), end="", file=f)

    def print_loc(self, prefix, loc, f):
        if loc["module"] == "__main__":
            print("%s* Line %d: " % (prefix, loc["line"]), end="", file=f)
        else:
            print("%s* Line %s/%d: " % (prefix, loc["module"], loc["line"]), end="", file=f)

    def path_str(self, path):
        result = path[0]
        for k in path[1:]:
            result += "[" + str(k) + "]"
        return result

    def value_eq(self, x, y):
        return x == y
        # if x["type"] != y["type"]:
        #     return False
        # return json.dumps(x) == json.dumps(y)   # TODO

    def value_lookup_list(self, path, value):
        assert path != []
        if path[0]["type"] != "int":
            return None
        idx = int(path[0]["value"])
        if idx >= 0 and idx < len(value):
            return self.value_lookup(path[1:], value[idx])
        return None

    def value_lookup_dict(self, path, value):
        assert path != []
        for kv in value:
            if self.value_eq(kv["key"], path[0]):
                return self.value_lookup(path[1:], kv["value"])
        return None

    def value_lookup(self, path, value):
        if path == []:
            return value
        if value["type"] == "list":
            return self.value_lookup_list(path, value["value"])
        if value["type"] == "dict":
            return self.value_lookup_dict(path, value["value"])
        return None

    def lookup(self, path, dict):
        # print("vl", path, dict)
        assert path != []
        assert path[0]["type"] == "atom"
        if path[0]["value"] not in dict:
            return None
        return self.value_lookup(path[1:], dict[path[0]["value"]])

    def listcmp(self, path, before, after, loc, f):
        alen = len(after["value"])
        blen = len(before["value"])
        if alen == blen:
            for i in range(blen):
                self.deepcmp(path + [i], before["value"][i], after["value"][i], loc, f)
        elif alen == blen + 1:
            for i in range(blen):
                self.deepcmp(path + [i], before["value"][i], after["value"][i], loc, f)
            self.print_loc("    ", loc, f)
            print("Set %s to %s" % (self.path_str(path + [blen]), verbose_string(after["value"][blen])), file=f)
        else:
            self.print_loc("    ", loc, f)
            print("Set %s to %s" % (self.path_str(path), verbose_string(after)), file=f)

    def dictcmp(self, path, before, after, loc, f):
        if all(kv["key"]["type"] == "atom" for kv in before["value"]) and all(kv["key"]["type"] == "atom" for kv in after["value"]):
            bd = { verbose_string(kv["key"]):kv["value"] for kv in before["value"] }
            ad = { verbose_string(kv["key"]):kv["value"] for kv in after["value"] }
            self.deepdiff(path, bd, ad, loc, f)
            return
        if all(kv["key"]["type"] == "int" for kv in before["value"]) and all(kv["key"]["type"] == "int" for kv in after["value"]):
            bd = { verbose_string(kv["key"]):kv["value"] for kv in before["value"] }
            ad = { verbose_string(kv["key"]):kv["value"] for kv in after["value"] }
            self.deepdiff(path, bd, ad, loc, f)
            return
        self.print_loc("    ", loc, f)
        print("Set %s to %s" % (self.path_str(path), verbose_string(after)), file=f)

    def deepcmp(self, path, before, after, loc, f):
        if before != after:
            if before["type"] == after["type"] == "list":
                self.listcmp(path, before, after, loc, f)
            elif before["type"] == after["type"] == "dict":
                self.dictcmp(path, before, after, loc, f)
            else:
                self.print_loc("    ", loc, f)
                print("Set %s to %s" % (self.path_str(path), verbose_string(after)), file=f)

    def deepdiff(self, path, before, after, loc, f):
        bk = set(before.keys())
        ak = set(after.keys())
        if len(ak - bk) != 0:
            for k in sorted(ak - bk):
                self.print_loc("    ", loc, f)
                print("Initialize %s to %s" % (self.path_str(path + [k]), verbose_string(after[k])), file=f)
        if len(bk - ak) != 0:
            for k in sorted(bk - ak):
                self.print_loc("    ", loc, f)
                print("Delete variable %s" % self.path_str(path + [k]), file=f)
        for k in sorted(ak & bk):
            self.deepcmp(path + [k], before[k], after[k], loc, f)

    def print_about(self, f):
        for ctx in self.contexts:
            if ctx["tid"] == self.tid:
                if ctx["mode"] == "terminated":
                    print("    * **Thread terminated**", file=f)
                elif ctx["mode"] != "failed":
                    mode = get_mode(ctx)
                    # if mode != "runnable":
                    #     print("    Mode = %s"%mode, file=f)
                    print("    * Preempted in ", end="", file=f)
                    verbose_print_trace(f, ctx["trace"])
                    if "next" in ctx:
                        self.about(ctx, "      ", f)

    def print_macrostep(self, f, mas):
        mis = mas["microsteps"]
        if mas["tid"] != self.tid:
            self.print_about(f)
            self.tid = mas["tid"]
            # print(file=f)
            # print("================================================", file=f)
            print("* Schedule thread T%s: "%mas["tid"], end="", file=f)
            # print("STATE(", self.shared, ")", file=f)
            ctx = mas["context"]
            trace = ctx["trace"]
            verbose_print_trace(f, trace)
            if False and len(trace) > 0:
                vars = trace[-1]["vars"]
                if len(vars) > 0:
                    print("method variables:", file=f)
                    for k, v in vars.items():
                        print("  %s: %s"%(k, verbose_string(v)), file=f)
            self.stack = [ verbose_string(v) for v in ctx["stack"] ]
            # if self.contexts != []:
            #     print("other threads:", file=f)
            #     self.dump_contexts(f, self.tid)
            if False and len(self.shared) != 0:
                print("    * Current values of global variables:", file=f)
                for k, v in self.shared.items():
                    print("        * %s: %s"%(k, verbose_string(v)), file=f)
            # print("================================================", file=f)
            self.lastmis = mis[0]
            self.start = int(self.lastmis["pc"])
            if False and "shared" in self.lastmis:
                self.shared = self.lastmis["shared"]
            lastpc = 0
            self.steps = ""
        self.contexts = mas["contexts"]
        for step in mis:
            self.step += 1
            pc = int(step["pc"])
            loc = self.locations[pc]

            if "interrupt" in step:
                self.print_loc("    ", loc, f)
                print("Interrupted: jump to interrupt handler first", file=f)
            elif self.code[pc]["op"] == "Store":
                self.print_loc("    ", loc, f)
                args = step["explain2"]["args"]
                # print("STORE(", self.shared, args, ")", file=f)
                if args == []:
                    print("Store: no args (pc=%d)???"%pc, file=f)
                else:
                    val = verbose_string(args[0])
                    assert args[1]["type"] == "address"
                    var = verbose_string(args[1])
                    oldval = self.lookup(args[1]["args"], self.shared)
                    if oldval == None:
                        print("Initialize %s to %s"%(var[1:], val), file=f)
                    elif self.value_eq(args[0], oldval):
                        print("Set %s to %s (unchanged)"%(var[1:], val), file=f)
                    else:
                        print("Set %s to %s (was %s)"%(var[1:], val, verbose_string(oldval)), file=f)
            elif "shared" in step and step["shared"] != self.shared:
                self.deepdiff([], self.shared, step["shared"], loc, f)
                # print("  Set global variables in line %d: " % loc["line"], end="", file=f)
                # verbose_print_vars(f, step["shared"])
            
            if "shared" in step:
                self.shared = step["shared"]

            if "choose" in step:
                self.print_loc("    ", loc, f)
                print("Choose %s"%verbose_string(step["choose"]), file=f)
            if "print" in step:
                self.print_loc("    ", loc, f)
                print("Print %s"%verbose_string(step["print"]), file=f)
            # if "trace" in step:
            #     print("  call trace:        ", end="", file=f)
            #     verbose_print_trace(f, step["trace"])
            # if "mode" in step and step["mode"] != "terminated":
            #     print("  New mode:          %s"%step["mode"], file=f)
            if "interruptlevel" in step:
                self.print_loc("    ", loc, f)
                if int(step["interruptlevel"]) == 0:
                    print("Interrupts enabled", file=f)
                elif int(step["interruptlevel"]) == 1:
                    print("Interrupts disabled", file=f)
                else:
                    print("Interrupt level: %s"%step["interruptlevel"], file=f)
            if "failure" in step:
                self.print_loc("    ", loc, f)
                if step["failure"].startswith("Harmony"):
                    print("%s"%step["failure"], file=f)
                else:
                    print("Operation failed: %s"%step["failure"], file=f)
            self.lastmis = step

    def run(self, outputfiles, top):
        with open(outputfiles["hco"], encoding='utf-8') as input:
            output = sys.stdout
            # with open("summary.txt", "w", encoding='utf-8') as output:
            print(file=output)
            print("----------------------------------------", file=output)
            print(file=output)
            assert isinstance(top, dict)
            if top["issue"] == "No issues":
                print("No issues were detected with this program.", file=output)
            else:
                if top["issue"] == "Safety violation":
                    print("**Summary: something went wrong in an execution**", file=output)
                elif top["issue"] == "Non-terminating state":
                    print("**Summary: some execution cannot terminate**", file=output)
                else:
                    print("**Summary: %s**" % top["issue"], file=output)

                assert isinstance(top["macrosteps"], list)
                self.hvm = top["hvm"]
                self.code = self.hvm["code"]

                print(file=output)
                print("Here is a summary of an execution that exhibits the issue:", file=output)
                print(file=output)
                self.locations = self.hvm["locs"]
                for mes in top["macrosteps"]:
                    self.print_macrostep(output, mes)
                self.print_about(output)

                if top["issue"] == "Non-terminating state":
                    print(file=output)
                    print("----------------------------------------", file=output)
                    print(file=output)
                    print("**Final state** (all threads have terminated or are blocked):", file=output)
                    print(file=output)
                    print("* Threads:", file=output)
                    self.dump_contexts(output, None)
                    print("* Variables:", file=output)
                    for k, v in self.shared.items():
                        print("    * %s: %s"%(k, verbose_string(v)), file=output)

                if len(contexts) > 0:
                    print(file=output)
                    print("This program uses the following contexts:", file=output)
                    print(file=output)
                    for i, ctx in enumerate(contexts):
                        print("* C%d:"%(i+1), file=output)
                        print_context(output, ctx)

                if len(self.hvm["modules"]) > 1:
                    print(file=output)
                    print("This program uses the following modules:", file=output)
                    print(file=output)
                    for modname in sorted(self.hvm["modules"].keys()):
                        mod = self.hvm["modules"][modname]
                        print("* %s: %s"%(modname, mod["file"]), file=output)



import json
import os
import queue
from pathlib import Path
from typing import Any, Tuple
from harmony_model_checker.harmony.jsonstring import json_string

class GenHTML:
    def __init__(self):
        self.top = {}
        self.nmegasteps = 0
        self.nmicrosteps = 0
        self.nthreads = 0
        self.vardir = {}
        self_dir = Path(__file__).parent
        self.style = (self_dir / "charm.css").read_text()
        self.js = (self_dir / "charm.js").read_text()
        self.timeWidth = 40

    def file_include(self, name, f):
        with open(name, encoding='utf-8') as g:
            print(g.read(), file=f)

    def html_megastep(self, step, tid, name, nmicrosteps, width, f):
        print("<tr id='mes%d'>"%(step-1), file=f)
        print("  <td align='right'>", file=f)
        print("    %d&nbsp;"%step, file=f)
        print("  </td>", file=f)

        print("  <td>", file=f)
        print("    T%s: %s"%(tid, name), file=f, end="")
        print("  </td>", file=f)

        print("  <td>", file=f)
        print("    <table border='0' style='max-width:%dpx; word-wrap: break-word;'>"%(self.timeWidth * 10), file=f)
        time = nmicrosteps
        nrows = (time + self.timeWidth - 1) // self.timeWidth
        print("      <tr><td><canvas id='timeline%d' width='%dpx' height='%dpx'>"%(step-1, self.timeWidth*10, 10*nrows), file=f)
        print("      </canvas></td></tr>", file=f)
        print("      <tr><td id='nextstep%d'></td></tr>"%(step-1), file=f)
        print("    </table>", file=f)
        print("  </td>", file=f)

        print("  <td align='center'>", file=f)
        print("  </td>", file=f)

        for i in range(width):
          print("  <td align='center'>", file=f)
          print("  </td>", file=f)

        print("  <td>", file=f)
        print("    <table id='log%d' border='1'>"%(step-1), file=f)
        print("    </table>", file=f)
        print("  </td>", file=f)
        print("</tr>", file=f)

    def vardim(self, d):
        if isinstance(d, dict):
            if d == {}:
                return (1, 0)
            totalwidth = 0
            maxheight = 0
            for k in sorted(d.keys()):
                (w, h) = self.vardim(d[k])
                totalwidth += w
                if h + 1 > maxheight:
                    maxheight = h + 1
            return (totalwidth, maxheight)
        else:
            return (1, 0)

    def varhdr(self, d: dict, name, nrows, f):
        q: 'queue.Queue[Tuple[dict, int]]' = queue.Queue()
        level = 0
        q.put((d, level))
        while not q.empty():
            (nd, nl) = q.get()
            if nl > level:
                print("</tr><tr>", file=f)
                level = nl
            if isinstance(nd, dict):
                for k in sorted(nd.keys()):
                    (w,h) = self.vardim(nd[k])
                    if k[0] == '"':
                        key = k[1:-1]
                    else:
                        key = k
                    if h == 0:
                        print("<td align='center' style='font-style: italic' colspan='%d' rowspan='%d'>%s</td>"%(w,nrows-nl,key), file=f)
                    else:
                        print("<td align='center' style='font-style: italic' colspan='%d'>%s</td>"%(w,key), file=f)
                    q.put((nd[k], nl+1))

    def html_top(self, f):
        if "macrosteps" not in self.top:
            print("<table border='1'>", file=f)
            print("  <thead>", file=f)
            print("    <tr>", file=f)
            print("      <th colspan='4' style='color:red;'>", file=f)
            print("        Issue:", self.top["issue"], file=f)
            print("      </th>", file=f)
            print("    </tr>", file=f)
            print("</table>", file=f)
            return

        (width, height) = self.vardim(self.vardir)
        print("<table border='1'>", file=f)
        print("  <thead>", file=f)
        print("    <tr>", file=f)
        print("      <th colspan='4' style='color:red;'>", file=f)
        print("        Issue:", self.top["issue"], file=f)
        # if "invpc" in self.top:
        #     print("        (Line %d)"%self.top["hvm"]["locs"][self.top["invpc"]]["line"], file=f)
        print("      </th>", file=f)
        print("      <th align='center' colspan='%d'>"%width, file=f)
        print("        Shared Variables", file=f)
        print("      </th>", file=f)
        print("      <th align='center' colspan='%d'>"%width, file=f)
        print("        Output", file=f)
        print("      </th>", file=f)
        print("    </tr>", file=f)

        print("    <tr>", file=f)
        print("      <th align='center' rowspan='%d'>"%height, file=f)
        print("        Turn", file=f)
        print("      </th>", file=f)
        print("      <th align='center' rowspan='%d'>"%height, file=f)
        print("        Thread", file=f)
        print("      </th>", file=f)
        print("      <th align='center' rowspan='%d'>"%height, file=f)
        print("        Instructions Executed", file=f)
        print("      </th>", file=f)
        print("      <th align='center' rowspan='%d'>"%height, file=f)
        print("        &nbsp;PC&nbsp;", file=f)
        print("      </th>", file=f)
        self.varhdr(self.vardir, "", height, f)
        print("    </tr>", file=f)
        print("  </thead>", file=f)

        print("  <tbody id='mestable'>", file=f)
        assert isinstance(self.top["macrosteps"], list)
        nsteps = 0
        tid = None
        name = None
        nmicrosteps = 0
        for mas in self.top["macrosteps"]:
            if tid == mas["tid"]:
                nmicrosteps += len(mas["microsteps"])
            else:
                if tid is not None:
                    self.html_megastep(nsteps, tid, name, nmicrosteps, width, f)
                nsteps += 1
                tid = mas["tid"]
                name = mas["name"]
                nmicrosteps = len(mas["microsteps"])
        self.html_megastep(nsteps, tid, name, nmicrosteps, width, f)
        print("  </tbody>", file=f)
        print("</table>", file=f)

    def html_botleft(self, f):
        print("<div id='table-wrapper'>", file=f)
        print("  <div id='table-scroll'>", file=f)
        print("    <table border='1'>", file=f)
        print("      <tbody>", file=f)
        alter = False
        pretty = self.top["hvm"]["pretty"]
        for pc, instr in enumerate(pretty):
            # if str(pc) in self.top["locations"]:  TODO: why was this here??
            alter = not alter
            print("        <tr id='P%d'>"%pc, file=f)
            print("          <td align='right'>", file=f)
            print("            <a name='P%d'>%d</a>&nbsp;"%(pc, pc), file=f)
            print("          </td>", file=f)
            print("          <td style='background-color: %s;'>"%("#E6E6E6" if alter else "white"), file=f)
            print("            <span title='%s' id='C%d'>"%(instr[1], pc), file=f)
            print("              %s"%instr[0], file=f)
            print("            </span>", file=f)
            print("          </td>", file=f)
            print("        </tr>", file=f)
        print("      </body>", file=f)
        print("    </table>", file=f)
        print("  </div>", file=f)
        print("</div>", file=f)

    # output a filename of f1 relative to f2
    def pathdiff(self, f1, f2):
        return os.path.relpath(f1, start=os.path.dirname(f2))

    def html_botright(self, f, outputfiles):
        if self.nthreads == 0:
            png = outputfiles["png"]
            if png is not None:
                if png[0] == "/":
                    print("      <img src='%s' alt='DFA image'>"%png, file=f)
                else:
                    assert outputfiles["htm"] is not None
                    print("      <img src='%s' alt='DFA image'>"%self.pathdiff(png, outputfiles["htm"]), file=f)
            return
        print("<table border='1'", file=f)
        print("  <thead>", file=f)
        print("    <tr>", file=f)
        print("      <th colspan='4'>Threads</th>", file=f)
        print("    </tr>", file=f)
        print("    <tr>", file=f)
        print("      <th>", file=f)
        print("        ID", file=f)
        print("      </th>", file=f)
        print("      <th>", file=f)
        print("        Status", file=f)
        print("      </th>", file=f)
        print("      <th>", file=f)
        print("        Stack Trace", file=f)
        print("      </th>", file=f)
        print("      <th>", file=f)
        print("        Stack Top", file=f)
        print("      </th>", file=f)
        print("    </tr>", file=f)
        print("  </thead>", file=f)
        print("  <tbody id='threadtable'>", file=f)
        maxtid = 0
        for i in range(self.nthreads):
            print("    <tr id='thread%d'>"%i, file=f)
            print("      <td align='center'>", file=f)
            print("        T%d"%i, file=f)
            print("      </td>", file=f)
            print("      <td align='center'>", file=f)
            print("        init", file=f)
            print("      </td>", file=f)
            print("      <td>", file=f)
            print("        <table id='threadinfo%d' border='1'>"%i, file=f)
            print("        </table>", file=f)
            print("      </td>", file=f)
            print("      <td align='left'>", file=f)
            print("      </td>", file=f)
            print("    </tr>", file=f)
        print("  </tbody>", file=f)
        print("</table>", file=f)

    def html_outer(self, f, outputfiles):
        print("<table>", file=f)
        print("  <tr>", file=f)
        print("    <td colspan='2'>", file=f)
        self.html_top(f)
        print("    </td>", file=f)
        print("  </tr>", file=f)
        print("  <tr><td></td></tr>", file=f)
        print("  <tr>", file=f)
        print("    <td colspan='2'>", file=f)
        print("      <h3 style='color:blue;'>", file=f)
        print("        <div id='coderow'>", file=f)
        print("        </div>", file=f)
        print("      </h3>", file=f)
        print("    </td>", file=f)
        print("  </tr>", file=f)
        print("  <tr><td></td></tr>", file=f)
        print("  <tr>", file=f)
        print("    <td valign='top'>", file=f)
        self.html_botleft(f)
        print("    </td>", file=f)
        print("    <td valign='top'>", file=f)
        self.html_botright(f, outputfiles)
        print("    </td>", file=f)
        print("  </tr>", file=f)
        print("</table>", file=f)

    def vardir_dump(self, d, path, index, f):
        if isinstance(d, dict) and d != {}:
            for k in sorted(d.keys()):
                index = self.vardir_dump(d[k], path + [k], index, f)
            return index
        if index > 0:
            print(",", file=f)
        print("  " + str(path), end="", file=f)
        return index + 1

    def html_script(self, f, outputfiles):
        print("<script>", file=f)
        print("var nthreads = %d;"%self.nthreads, file=f)
        print("var nmegasteps = %d;"%self.nmegasteps, file=f)
        print("var vardir = [", file=f)
        self.vardir_dump(self.vardir, [], 0, f)
        print(file=f)
        print("];", file=f)
        print("var state =", file=f)
        self.file_include(outputfiles["hco"], f)
        print(";", file=f)
        print(self.js, file=f)
        # file_include("charm.js", f)
        print("</script>", file=f)

    def html_body(self, f, outputfiles):
        print("<body>", file=f)
        self.html_outer(f, outputfiles)
        self.html_script(f, outputfiles)
        print("</body>", file=f)

    def html_head(self, f):
        print("<head>", file=f)
        print("  <meta charset='UTF-8'></meta>", file=f)
        print("  <style>", file=f)
        print(self.style, file=f)
        print("  </style>", file=f)
        print("</head>", file=f)

    def html(self, f, outputfiles):
        print("<html>", file=f)
        self.html_head(f)
        self.html_body(f, outputfiles)
        print("</html>", file=f)

    def var_convert(self, v):
        if v["type"] != "dict":
            return json_string(v)
        for kv in v["value"]:
            k = kv["key"]
            if k["type"] != "atom" and k["type"] != "int" and k["type"] != "bool":
                return json_string(v)
        d = {}
        for kv in v["value"]:
            k = json_string(kv["key"])
            d[k] = self.var_convert(kv["value"])
        return d

    def dict_merge(self, vardir, d):
        for (k, v) in d.items():
            if not isinstance(v, dict):
                vardir[k] = v
            else:
                if k not in vardir:
                    vardir[k] = {}
                elif not isinstance(vardir[k], dict):
                    continue
                self.dict_merge(vardir[k], v)

    def vars_add(self, vardir, shared):
        d = {}
        for (k, v) in shared.items():
            val = self.var_convert(v)
            # if val != {}:
            d[k] = val
        self.dict_merge(vardir, d)

    def run(self, outputfiles):
        # First figure out how many megasteps there are and how many threads
        lasttid = -1
        with open(outputfiles["hco"], encoding='utf-8') as f:
            self.top = json.load(f, strict=False)
            assert isinstance(self.top, dict)
            if "macrosteps" in self.top:
                macrosteps = self.top["macrosteps"]
                for mas in macrosteps:
                    tid = int(mas["tid"])
                    if tid >= self.nthreads:
                        self.nthreads = tid + 1
                    if tid != lasttid:
                        self.nmegasteps += 1
                        lasttid = tid
                    self.nmicrosteps += len(mas["microsteps"])
                    for mis in mas["microsteps"]:
                        if "shared" in mis:
                            self.vars_add(self.vardir, mis["shared"])
                    for ctx in mas["contexts"]:
                        tid = int(ctx["tid"])
                        if tid >= self.nthreads:
                            self.nthreads = tid + 1

        with open(outputfiles["htm"], "w", encoding='utf-8') as out:
            self.html(out, outputfiles)



import sys
import os
import getopt
import traceback
import collections
import time
import math


def islower(c):
    return c in "abcdefghijklmnopqrstuvwxyz"

def isupper(c):
    return c in "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

def isletter(c):
    return islower(c) or isupper(c)

def isnumeral(c):
    return c in "0123456789"

def isalnum(c):
    return isletter(c) or isnumeral(c)

def isnamechar(c):
    return isalnum(c) or c == "_"

def isnumber(s):
    return all(isnumeral(c) for c in s)

reserved = {
    "abs",
    "all",
    "and",
    "any",
    "as",
    "assert",
    "atLabel",
    "atomic",
    "atomically",
    "await",
    "bag",
    "choose",
    "const",
    "countLabel",
    "def",
    "del",
    "dict",
    "elif",
    "else",
    "end",
    "eternal",
    "exists",
    "from",
    "False",
    "finally",
    "for",
    "global",
    "go",
    "hash",
    "if",
    "import",
    "in",
    "inf",
    "invariant",
    "keys",
    "lambda",
    "len",
    "let",
    "max",
    "min",
    "mod",
    "None",
    "not",
    "or",
    "pass",
    "print",
    "returns",
    "save",
    "select",
    "sequential",
    "setintlevel",
    "spawn",
    "stop",
    "str",
    "this",
    "trap",
    "True",
    "type",
    "var",
    "when",
    "where",
    "while"
}

def isreserved(s: str):
    return s in reserved

def isname(s):
    return (not isreserved(s)) and (isletter(s[0]) or s[0] == "_") and \
                    all(isnamechar(c) for c in s)

def isunaryop(s):
    return s in [ "^", "-", "atLabel", "bagsize", "cardinality",
            "min", "max", "nametag", "not", "keys", "len", "processes" ]

def isbinaryop(s):
    return s in [
        "==", "!=", "..", "in", "and", "or",
        "-", "+", "*", "/", "%", "<", "<=", ">", ">="
    ];

tokens = [ "==", "!=", "<=", ">=", "..", "->", "[]" ]

def putchar(c, fd):
    print(c, end="", file=fd)

def capitalize(s):
    return "".join([ x[0].upper() + x[1:] for x in s.split("-")])

def nextLine(s, fd, pmap):
    while s != "":
        # skip over line comments
        if s.startswith("#"):
            s = s[1:]
            print("\\emph{\\#", end="", file=fd)
            while len(s) > 0 and s[0] != '\n':
                if s[0] in ["&", "%", "{", "}", "#", "^", "_"]:
                    print("\\" + s[0], end="", file=fd)
                elif s[0] in [ "<", ">", "-" ]:
                    print("$" + s[0] + "$", end="", file=fd)
                else:
                    putchar(s[0], fd)
                s = s[1:]
            print("}", end="", file=fd)
            continue

        # see if it's a multi-character token.  Match with the longest one
        found = ""
        for t in tokens:
            if s.startswith(t) and len(t) > len(found):
                found = t
        if found != "":
            if found in [ "<=", ">=" ]:
                print("$%s$"%found, end="", file=fd)
            elif found == "->":
                print("$\\rightarrow$", end="", file=fd)
            elif found == "[]":
                print("[\,]", end="", file=fd)
            else:
                print(found, end="", file=fd)
            s = s[len(found):]
            continue

        # see if a sequence of letters and numbers
        if isnamechar(s[0]):
            i = 0
            while i < len(s) and isnamechar(s[i]):
                i += 1
            found = s[:i]
            if found in reserved:
                nt = "reserved"
            elif isnumeral(found[0]):
                nt = "number"
            elif found in pmap:
                (nt, _) = pmap[found]
            else:
                nt = "var"
            found = found.replace("_", "\\_")
            print("\\hny%s{%s}"%(capitalize(nt),found), end="", file=fd)
            s = s[i:]
            continue

        # string
        if s[0] == '"':
            print('\\texttt{"}', end="", file=fd)
            i = 1
            while i < len(s) and s[i] != '"':
                if s[i] in ["&", "%", "{", "}", "#", "^", "_"]:
                    print("\\" + s[i], end="", file=fd)
                elif s[i] in [ "<", ">", "-" ]:
                    print("$" + s[i] + "$", end="", file=fd)
                else:
                    putchar(s[i], fd)
                i += 1
            if i < len(s):
                i += 1
            print('\\texttt{"}', end="", file=fd)
            s = s[i:]
            continue

        if s[0] == "'":
            print("{\\textquotesingle}", end="")
        elif s[0] in ["&", "%", "{", "}"]:
            print("\\" + s[0], end="", file=fd)
        elif s[0] == "~":
            print("\\string~", end="", file=fd)
        elif s[0] == "^":
            print("\\^{}", end="", file=fd)
        elif s[0] == "|":
            print("$\\vert$", end="", file=fd)
        else:
            if s[0] == "-":
                print("--", end="", file=fd)
            elif s[0] in [ "<", ">" ]:
                print("$" + s[0] + "$", end="", file=fd)
            else:
                putchar(s[0], fd)
        s = s[1:]

def lexer(s, file, fd, pmap):
    line = 1
    column = 1
    while s != "":
        if column == 1:
            if line != 1: print("\\\\\n", file=fd)
            print("\\>\\hnyLine{%d}\\>"%line, end="", file=fd)

        # see if it's a blank
        if s[0] in { " ", "\t" }:
            s = s[1:]
            column += 1
            print("{\\hnySp}", end="", file=fd)
            continue

        # skip over nested comments
        if s.startswith("(*"):
            count = 1
            s = s[2:]
            column += 2
            while count != 0 and s != "":
                if s.startswith("(*"):
                    count += 1
                    s = s[2:]
                    column += 2
                elif s.startswith("*)"):
                    count -= 1
                    s = s[2:]
                    column += 2
                elif s[0] == "\n":
                    s = s[1:]
                    line += 1
                    column = 1
                else:
                    s = s[1:]
                    column += 1
            continue

        # Find the end of the line, if any
        i = s.find("\n")
        if i < 0:
            nextLine(s, fd, pmap)
            s = ""
        else:
            nextLine(s[:i], fd, pmap)
            s = s[i+1:]
            line += 1
            column = 1

# TODO.  Can this be done efficiently in Python with just read()?
def loadAll(filename):
    all = ""
    with open(filename) as fd:
        for line in fd:
            all += line
    return all

def doProcess(filename, fd, pmap):
    all = loadAll(filename)
    print("% \\documentclass{article}", file=fd)
    print("% \\begin{document}", file=fd)
    print("\\providecommand{\\hnyLine}[1]{{\\tiny #1}\\'}", file=fd)
    print("\\providecommand{\\hnyReserved}[1]{\\textbf{#1}}", file=fd)
    print("\\providecommand{\\hnyModule}[1]{\\texttt{#1}}", file=fd)
    print("\\providecommand{\\hnyConstant}[1]{\\texttt{#1}}", file=fd)
    print("\\providecommand{\\hnyImported}[1]{\\texttt{#1}}", file=fd)
    print("\\providecommand{\\hnyGlobal}[1]{\\textit{#1}}", file=fd)
    print("\\providecommand{\\hnyLocalVar}[1]{\\textit{#1}}", file=fd)
    print("\\providecommand{\\hnyLocalConst}[1]{\\textit{#1}}", file=fd)
    print("\\providecommand{\\hnyVar}[1]{\\textit{#1}}", file=fd)
    print("\\providecommand{\\hnyNumber}[1]{#1}", file=fd)
    print("\\providecommand{\\hnySp}{\\rule{.3em}{0em}}", file=fd)
    print("\\begin{tabbing}", file=fd)
    print("X\\=XX\\=XXX\\kill", file=fd)
    lexer(all, filename, fd, pmap)
    print("\\end{tabbing}", file=fd)
    print("% \\end{document}", file=fd)

def tex_main(f, code, scope):
    doProcess(scope.file, f, scope.pmap)



from typing import Union, Optional
from harmony_model_checker.harmony.AbstractASTVisitor import AbstractASTVisitor
from harmony_model_checker.harmony.ast import *


class DumpASTVisitor(AbstractASTVisitor):

    def __init__(self, indent_unit: Union[int, str] = 2):
        if isinstance(indent_unit, int):
            self.indent_unit = " " * indent_unit
        else:
            self.indent_unit = indent_unit

    def __call__(self, ast: AST, depth: Optional[int] = None):
        if depth is None:
            self.buffer = ""
            ast.accept_visitor(self, depth=0)
            return self.buffer
        else:
            ast.accept_visitor(self, depth=depth)
        
    def p(self, x, indent=0):
        self.buffer += self.indent_unit * indent + str(x) + "\n"

    def _visit_comprehension_iter(self, iter, value, depth):
        for type, rest in iter:
            if type == "for":
                _, _, expr = rest
                self(expr, depth + 1)
            
            elif type == "where":
                self(rest, depth + 1)

        self(value, depth + 1)

    def visit_constant(self, node: ConstantAST, depth):
        self.p(f"Constant {node.token}", indent=depth)

    def visit_name(self, node: NameAST, depth):
        self.p(f"Name {node.token}", indent=depth)

    def visit_set(self, node: SetAST, depth):
        self.p(f"Set {node.token}", indent=depth)

        for subnode in node.collection:
            self(subnode, depth + 1)
    
    def visit_range(self, node: RangeAST, depth):
        self.p(f"Range {node.token}", indent=depth)

        self(node.lhs, depth + 1)
        self(node.rhs, depth + 1)

    def visit_tuple(self, node: TupleAST, depth):
        self.p(f"Tuple {node.token}", indent=depth)

        for subnode in node.list:
            self(subnode, depth + 1)

    def visit_dict(self, node: DictAST, depth):
        self.p(f"Dict {node.token}", indent=depth)

        for (k, v) in node.record:
            self(k, depth + 1)
            self(v, depth + 1)

    def visit_set_comprehension(self, node: SetComprehensionAST, depth):
        self.p(f"SetComprehension {node.token}", indent=depth)

        self._visit_comprehension_iter(node.iter, node.value, depth)

    def visit_dict_comprehension(self, node: DictComprehensionAST, depth):
        self.p(f"DictComprehension {node.token}", indent=depth)

        self._visit_comprehension_iter(node.iter, node.value, depth)

    def visit_list_comprehension(self, node: ListComprehensionAST, depth):
        self.p(f"ListComprehension {node.token}", indent=depth)

        self._visit_comprehension_iter(node.iter, node.value, depth)

    def visit_nary(self, node: NaryAST, depth):
        self.p(f"Nary {node.token}", indent=depth)

        for subnode in node.args:
            self(subnode, depth + 1)

    def visit_cmp(self, node: CmpAST, depth):
        self.p(f"Cmp {node.token}", indent=depth)

        for subnode in node.args:
            self(subnode, depth + 1)

    def visit_apply(self, node: ApplyAST, depth):
        self.p(f"Apply {node.token}", indent=depth)

        self(node.method, depth + 1)
        self(node.arg, depth + 1)

    def visit_pointer(self, node: PointerAST, depth):
        self.p(f"Pointer {node.token}", indent=depth)

        self(node.expr, depth + 1)

    def visit_assignment(self, node: AssignmentAST, depth):
        self.p(f"Assignment {node.token}", indent=depth)

        for lhs in node.lhslist:
            self(lhs, depth + 1)

        self(node.rv, depth + 1)

    def visit_del(self, node: DelAST, depth):
        self.p(f"Del {node.token}", indent=depth)

    def visit_set_int_level(self, node: SetIntLevelAST, depth):
        self.p(f"SetIntLevel {node.token}", indent=depth)

    def visit_save(self, node: SaveAST, depth):
        self.p(f"Save {node.token}", indent=depth)

    def visit_stop(self, node: StopAST, depth):
        self.p(f"Stop {node.token}", indent=depth)

    def visit_address(self, node: AddressAST, depth):
        self.p(f"Address {node.token}", indent=depth)

        self(node.lv, depth + 1)

    def visit_pass(self, node: PassAST, depth):
        self.p(f"Pass {node.token}", indent=depth)

    def visit_block(self, node: BlockAST, depth):
        self.p(f"Block {node.token}", indent=depth)

        for subnode in node.b:
            self(subnode, depth + 1)

    def visit_if(self, node: IfAST, depth):
        self.p(f"If {node.token}", indent=depth)

        for alt in node.alts:
            _, stat, _, _ = alt 
            self(stat, depth + 1)

        if node.stat is not None:
            self(node.stat, depth + 1)

    def visit_while(self, node: WhileAST, depth):
        self.p(f"While {node.token}", indent=depth)

        self(node.cond, depth + 1)
        self(node.stat, depth + 1)

    def visit_invariant(self, node: InvariantAST, depth):
        self.p(f"Invariant {node.token}", indent=depth)

        self(node.cond, depth + 1)

    def visit_let(self, node: LetAST, depth):
        self.p(f"Let {node.token}", indent=depth)

        for _, expr in node.vars:
            self(expr, depth + 1)

        self(node.stat, depth + 1)

    def visit_var(self, node: VarAST, depth):
        self.p(f"Var {node.token}", indent=depth)

        for _, expr in node.vars:
            self(expr, depth + 1)

    def visit_for(self, node: ForAST, depth):
        self.p(f"For {node.token}", indent=depth)

        self._visit_comprehension_iter(node.iter, node.value, depth)
    
    def visit_let_when(self, node: LetWhenAST, depth):
        self.p(f"LetWhen {node.token}", indent=depth)

        for var_or_cond in node.vars_and_conds:
            if var_or_cond[0] == 'var':
                _, expr = var_or_cond[1:]
                self(expr, depth + 1)

            elif var_or_cond[0] == 'cond':
                cond = var_or_cond[1]
                self(cond, depth + 1)
        
    def visit_atomic(self, node: AtomicAST, depth):
        self.p(f"Atomic {node.token}", indent=depth)

        self(node.stat, depth + 1)

    def visit_assert(self, node: AssertAST, depth):
        self.p(f"Assert {node.token}", indent=depth)

        self(node.cond, depth + 1)

        if node.expr is not None:
            self(node.expr, depth + 1)

    def visit_print(self, node: PrintAST, depth):
        self.p(f"Print {node.token}", indent=depth)

        self(node.expr, depth + 1)

    def visit_method(self, node: MethodAST, depth):
        self.p(f"Method {node.token}", indent=depth)

        self(node.stat, depth + 1)

    def visit_lambda(self, node: LambdaAST, depth):
        self.p(f"Lambda {node.token}", indent=depth)

        self(node.stat, depth + 1)

    def visit_call(self, node: CallAST, depth):
        self.p(f"Call {node.token}", indent=depth)

        self(node.expr, depth + 1)

    def visit_spawn(self, node: SpawnAST, depth):
        self.p(f"Spawn {node.token}", indent=depth)

        self(node.method, depth + 1)
        self(node.arg, depth + 1)

    def visit_trap(self, node: TrapAST, depth):
        self.p(f"Trap {node.token}", indent=depth)

        self(node.arg, depth + 1)
        self(node.method, depth + 1)

    def visit_go(self, node: GoAST, depth):
        self.p(f"Go {node.token}", indent=depth)

    def visit_import(self, node: ImportAST, depth):
        self.p(f"Import {node.token}", indent=depth)

    def visit_from(self, node: FromAST, depth):
        self.p(f"From {node.token}", indent=depth)

    def visit_location(self, node: LocationAST, depth):
        self.p(f"Location {node.token}", indent=depth)

        self(node.ast, depth + 1)

    def visit_label_stat(self, node: LabelStatAST, depth):
        self.p(f"LabelStat {node.token}", indent=depth)

    def visit_sequential(self, node: SequentialAST, depth):
        self.p(f"Sequential {node.token}", indent=depth)

    def visit_const(self, node: ConstAST, depth):
        self.p(f"Const {node.token}", indent=depth)



from typing import Any, Union
from antlr4.Token import CommonToken  # type: ignore

from harmony_model_checker.parser.HarmonyVisitor import HarmonyVisitor
from harmony_model_checker.parser.HarmonyParser import HarmonyParser
from harmony_model_checker.harmony.ast import *

class HarmonyVisitorImpl(HarmonyVisitor):

    def __init__(self, file: str):
        assert isinstance(file, str)
        self.file = file
        self.associative_operators = { "+", "*", "|", "&", "^", "and", "or" }

    def get_token(self, tkn: Union[CommonToken, tuple], value: Any) -> tuple:
        if isinstance(tkn, CommonToken):
            line = tkn.line
            col = tkn.column + 1
        else:
            line = tkn[2]
            col = tkn[3]

        assert isinstance(col, int)
        assert isinstance(line, int)

        return value, self.file, line, col

    # Visit a parse tree produced by HarmonyParser#import_stmt.
    def visitImport_stmt(self, ctx: HarmonyParser.Import_stmtContext):
        if ctx.import_from():
            return self.visit(ctx.import_from())
        elif ctx.import_name():
            return self.visit(ctx.import_name())
        tkn = self.get_token(ctx.start, ctx.start.text)
        raise HarmonyCompilerError(
            message="Failed to parse import",
            filename=self.file,
            line=tkn[2],
            column=tkn[3],
            lexeme=tkn[0]
        )

    # Visit a parse tree produced by HarmonyParser#import_name.
    def visitImport_name(self, ctx: HarmonyParser.Import_nameContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return ImportAST(endtoken, tkn, False, self.visit(ctx.import_names_seq()))

    # Visit a parse tree produced by HarmonyParser#import_from.
    def visitImport_from(self, ctx: HarmonyParser.Import_fromContext):
        dotted_name = self.get_token(ctx.NAME().symbol, str(ctx.NAME()))
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if ctx.STAR():
            return FromAST(endtoken, tkn, False, dotted_name, [])
        names = self.visit(ctx.import_names_seq())
        return FromAST(endtoken, tkn, False, dotted_name, names)
    
    def visitImport_names_seq(self, ctx: HarmonyParser.Import_names_seqContext):
        return [self.get_token(d.symbol, str(d)) for d in ctx.NAME()]

    # Visit a parse tree produced by HarmonyParser#program.
    def visitProgram(self, ctx: HarmonyParser.ProgramContext):
        stmts = [s for stmt in ctx.stmt() for s in self.visit(stmt) if s is not None]
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if len(stmts) == 0:
            raise HarmonyCompilerError(
                message="Empty program",
                filename=self.file,
                line=tkn[2],
                column=tkn[3],
                lexeme=tkn[0]
            )
        return BlockAST(endtoken, tkn, False, stmts, endtoken)

    # Visit a parse tree produced by HarmonyParser#tuple_bound.
    def visitTuple_bound(self, ctx:HarmonyParser.Tuple_boundContext):
        if ctx.bound() is not None:
            return self.visit(ctx.bound())
        if ctx.NAME():
            name = ctx.NAME()
            return self.get_token(name.symbol, str(name))
        return []

    # Visit a parse tree produced by HarmonyParser#bound.
    def visitBound(self, ctx: HarmonyParser.BoundContext):
        bv = [self.visit(t) for t in ctx.tuple_bound()]
        if len(bv) == 1:
            return bv[0]
        return bv

    # Visit a parse tree produced by HarmonyParser#comp_op.
    def visitComp_op(self, ctx: HarmonyParser.Comp_opContext):
        return self.get_token(ctx.start, ctx.getText())

    # Visit a parse tree produced by HarmonyParser#unary_op.
    def visitUnary_op(self, ctx: HarmonyParser.Unary_opContext):
        return self.get_token(ctx.start, ctx.getText())

    # Visit a parse tree produced by HarmonyParser#int.
    def visitInt(self, ctx: HarmonyParser.IntContext):
        txt = ctx.getText()
        tkn = self.get_token(ctx.start, int(txt, 0))
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return ConstantAST(endtoken, tkn)

    # Visit a parse tree produced by HarmonyParser#bool.
    def visitBool(self, ctx: HarmonyParser.BoolContext):
        txt = ctx.getText()
        tkn = self.get_token(ctx.start, txt == 'True')
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return ConstantAST(endtoken, tkn)

    # Visit a parse tree produced by HarmonyParser#atom.
    def visitAtom(self, ctx: HarmonyParser.AtomContext):
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        lexeme = ctx.getText()
        if lexeme.startswith("0x"):
            v = chr(int(lexeme, 16))
            tkn = self.get_token(ctx.start, v)
            return ConstantAST(endtoken, tkn)
        else:
            tkn = self.get_token(ctx.start, lexeme[1:])
            return ConstantAST(endtoken, tkn)

    # Visit a parse tree produced by HarmonyParser#name.
    def visitName(self, ctx: HarmonyParser.NameContext):
        tkn = self.get_token(ctx.start, ctx.getText())
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return NameAST(endtoken, tkn)

    # Visit a parse tree produced by HarmonyParser#str.
    def visitStr(self, ctx: HarmonyParser.StrContext):
        value = ctx.getText()
        if value.startswith("'''") and value.endswith("'''"):
            s = value[3:-3]
        elif value.startswith('"""') and value.endswith('"""'):
            s = value[3:-3]
        elif value.startswith("'") and value.endswith("'"):
            s = value[1:-1]
        elif value.startswith('"') and value.endswith('"'):
            s = value[1:-1]
        else:
            tkn = self.get_token(ctx.start, value)
            raise HarmonyCompilerError(
                message="Unable to parse string",
                filename=self.file,
                line=tkn[2],
                column=tkn[3],
                lexeme=tkn[0]
            )
        tkn = self.get_token(ctx.start, s)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return ConstantAST(endtoken, tkn)

    def visitPrint_stmt(self, ctx: HarmonyParser.Print_stmtContext):
        cond = self.visit(ctx.expr())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return PrintAST(endtoken, tkn, False, cond)

    # Visit a parse tree produced by HarmonyParser#none.
    def visitNone(self, ctx: HarmonyParser.NoneContext):
        tkn = self.get_token(ctx.start, AddressValue(None, []))
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return ConstantAST(endtoken, tkn)

    # Visit a parse tree produced by HarmonyParser#address.
    # def visitAddress(self, ctx: HarmonyParser.AddressContext):
    #     expr = self.visit(ctx.expr_rule())
    #     tkn = self.get_token(ctx.start, ctx.start.text)
    #     endtoken = self.get_token(ctx.stop, ctx.stop.text)
    #     return AddressAST(endtoken, tkn, expr)

    # Visit a parse tree produced by HarmonyParser#assign_op.
    def visitAssign_op(self, ctx: HarmonyParser.Assign_opContext):
        return self.get_token(ctx.start, ctx.getText())

    # Visit a parse tree produced by HarmonyParser#aug_assign_op.
    def visitAug_assign_op(self, ctx: HarmonyParser.Aug_assign_opContext):
        return self.get_token(ctx.start, ctx.getText())

    # Visit a parse tree produced by HarmonyParser#expr_stmt.
    def visitExpr_stmt(self, ctx: HarmonyParser.Expr_stmtContext):
        e = self.visit(ctx.expr_rule())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return CallAST(endtoken, tkn, False, e)

    # Visit a parse tree produced by HarmonyParser#assign_stmt.
    def visitAssign_stmt(self, ctx: HarmonyParser.Assign_stmtContext):
        ops = [self.visit(e) for e in ctx.assign_op()]
        expr = [self.visit(e) for e in ctx.tuple_rule()]
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return AssignmentAST(endtoken, tkn, expr[:-1], expr[-1], ops, False)

    # Visit a parse tree produced by HarmonyParser#aug_assign_stmt.
    def visitAug_assign_stmt(self, ctx: HarmonyParser.Aug_assign_stmtContext):
        op = self.visit(ctx.aug_assign_op())
        expressions = [self.visit(e) for e in ctx.tuple_rule()]
        lhs = expressions[0]
        rhs = expressions[1]
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return AuxAssignmentAST(endtoken, tkn, lhs, rhs, op, False)

    # Visit a parse tree produced by HarmonyParser#const_assign_stmt.
    def visitConst_assign_stmt(self, ctx: HarmonyParser.Const_assign_stmtContext):
        lvalues = self.visit(ctx.bound())
        value = self.visit(ctx.expr())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return ConstAST(endtoken, tkn, False, lvalues, value)

    # Visit a parse tree produced by HarmonyParser#assert_stmt.
    def visitAssert_stmt(self, ctx: HarmonyParser.Assert_stmtContext):
        assertion = self.visit(ctx.expr(0))
        msg = self.visit(ctx.expr(1)) if ctx.expr(1) is not None else None
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return AssertAST(endtoken, tkn, False, assertion, msg)

    # Visit a parse tree produced by HarmonyParser#await_stmt.
    def visitAwait_stmt(self, ctx: HarmonyParser.Await_stmtContext):
        cond = self.visit(ctx.expr())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return LetWhenAST(endtoken, tkn, False, [('cond', cond, tkn, endtoken)], PassAST(endtoken, tkn, False))

    # Visit a parse tree produced by HarmonyParser#trap_stmt.
    def visitTrap_stmt(self, ctx: HarmonyParser.Trap_stmtContext):
        func = self.visit(ctx.expr())
        # if not isinstance(func, ApplyAST):
        #     tkn = self.get_token(func.token, func.token[0])
        #     raise HarmonyCompilerError(
        #         message="Expected a method call but found something else.",
        #         filename=self.file,
        #         line=tkn[2],
        #         column=tkn[3],
        #         lexeme=tkn[0]
        #     )
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return TrapAST(endtoken, tkn, False, func)

    # Visit a parse tree produced by HarmonyParser#pass_stmt.
    def visitPass_stmt(self, ctx: HarmonyParser.Pass_stmtContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return PassAST(endtoken, tkn, False)

    # Visit a parse tree produced by HarmonyParser#Finally_stmt.
    def visitFinally_stmt(self, ctx: HarmonyParser.Finally_stmtContext):
        cond = self.visit(ctx.expr())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return FinallyAST(endtoken, cond, tkn, False)

    # Visit a parse tree produced by HarmonyParser#invariant_stmt.
    def visitInvariant_stmt(self, ctx: HarmonyParser.Invariant_stmtContext):
        cond = self.visit(ctx.expr())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return InvariantAST(endtoken, cond, tkn, False)

    # Visit a parse tree produced by HarmonyParser#del_stmt.
    def visitDel_stmt(self, ctx: HarmonyParser.Del_stmtContext):
        e = self.visit(ctx.expr())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return DelAST(endtoken, tkn, False, e)

    def visitSimple_stmt(self, ctx: HarmonyParser.Simple_stmtContext):
        if ctx.ATOMICALLY():
            a = self.visit(ctx.children[-1])
            assert isinstance(a, AST)
            atom = ctx.ATOMICALLY()
            a.atomically = self.get_token(atom.getSymbol(), atom.getText())
            return a
        return super().visitSimple_stmt(ctx)

    def visitCompound_stmt(self, ctx: HarmonyParser.Compound_stmtContext):
        if ctx.ATOMICALLY():
            a = self.visit(ctx.children[-1])
            assert isinstance(a, AST)
            atom = ctx.ATOMICALLY()
            a.atomically = self.get_token(atom.getSymbol(), atom.getText())
            return a
        return super().visitCompound_stmt(ctx)

    # Visit a parse tree produced by HarmonyParser#spawn_stmt.
    def visitSpawn_stmt(self, ctx: HarmonyParser.Spawn_stmtContext):
        is_eternal = ctx.ETERNAL() is not None
        target = self.visit(ctx.expr())
        # if not isinstance(target, ApplyAST):
        #     tkn = self.get_token(target.token, target.token[0])
        #     raise HarmonyCompilerError(
        #         message="Expected a method call but found something else.",
        #         filename=self.file,
        #         line=tkn[2],
        #         column=tkn[3],
        #         lexeme=tkn[0]
        #     )
        # method, arg = target.method, target.arg
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        # return SpawnAST(endtoken, tkn, False, method, arg, None, is_eternal)
        return SpawnAST(endtoken, tkn, False, target, None, is_eternal)

    # Visit a parse tree produced by HarmonyParser#go_stmt.
    def visitGo_stmt(self, ctx: HarmonyParser.Go_stmtContext):
        target = self.visit(ctx.expr(0))
        args = self.visit(ctx.expr(1))
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return GoAST(endtoken, tkn, False, target, args)

    # Visit a parse tree produced by HarmonyParser#sequential_stmt.
    def visitSequential_stmt(self, ctx: HarmonyParser.Sequential_stmtContext):
        expr = [self.visit(e) for e in ctx.expr()]
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return SequentialAST(endtoken, tkn, False, expr)

    # Visit a parse tree produced by HarmonyParser#global_stmt.
    def visitGlobal_stmt(self, ctx: HarmonyParser.Global_stmtContext):
        expr = [self.visit(e) for e in ctx.expr()]
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return GlobalAST(endtoken, tkn, False, expr)

    # Visit a parse tree produced by HarmonyParser#builtin_stmt.
    def visitBuiltin_stmt(self, ctx: HarmonyParser.Builtin_stmtContext):
        name = str(ctx.NAME())
        nametkn = self.get_token(ctx.NAME().symbol, name)
        value = str(ctx.STRING())[1:-1]     # remove quotes
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return BuiltinAST(endtoken, tkn, NameAST(nametkn, nametkn), value)

    # Visit a parse tree produced by HarmonyParser#atomic_block.
    def visitAtomic_block(self, ctx: HarmonyParser.Atomic_blockContext):
        stmts = self.visit(ctx.block())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        atom = ctx.ATOMICALLY()
        atoken = self.get_token(atom.getSymbol(), atom.getText())
        colon = ctx.COLON()
        ctoken = self.get_token(colon.getSymbol(), colon.getText())
        return AtomicAST(endtoken, tkn, atoken, stmts, ctoken)

    # Visit a parse tree produced by HarmonyParser#for_block.
    def visitFor_block(self, ctx: HarmonyParser.For_blockContext):
        iterables = self.visit(ctx.iter_parse())
        stmts = self.visit(ctx.block())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return ForAST(endtoken, iterables, stmts, tkn, False)

    # Visit a parse tree produced by HarmonyParser#let_decl.
    def visitLet_decl(self, ctx: HarmonyParser.Let_declContext):
        lvalues = self.visit(ctx.bound())
        values = self.visit(ctx.tuple_rule())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        op = self.visit(ctx.assign_op())
        return "var", lvalues, values, tkn, endtoken, op

    # Visit a parse tree produced by HarmonyParser#when_decl.
    def visitWhen_decl(self, ctx:HarmonyParser.When_declContext):
        expr = self.visit(ctx.expr())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if ctx.EXISTS() is not None:
            bv = self.visit(ctx.bound())
            return 'exists', bv, expr, tkn, endtoken
        else:
            return 'cond', expr, tkn, endtoken

    # Visit a parse tree produced by HarmonyParser#let_when_decl.
    def visitLet_when_decl(self, ctx:HarmonyParser.Let_when_declContext):
        decl = []
        if ctx.when_decl() is not None:
            decl.append(self.visit(ctx.when_decl()))
        elif ctx.let_decl() is not None:
            decl.append(self.visit(ctx.let_decl()))
        else:
            tkn = self.get_token(ctx.start, ctx.start.text)
            raise HarmonyCompilerError(
                message="Unexpected declaration in let/when block declaration.",
                filename=self.file,
                line=tkn[2],
                column=tkn[3],
                lexeme=tkn[0]
            )
        if ctx.let_when_decl() is not None:
            decl.extend(self.visit(ctx.let_when_decl()))
        return decl

    # Visit a parse tree produced by HarmonyParser#let_when_block.
    def visitLet_when_block(self, ctx:HarmonyParser.Let_when_blockContext):
        vars_and_conds = self.visit(ctx.let_when_decl())
        stmts = self.visit(ctx.block())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if all(vac[0] == 'var' for vac in vars_and_conds):
            vars = [vac[1:] for vac in vars_and_conds if vac[0] == 'var']
            return LetAST(endtoken, tkn, False, vars, stmts)
        else:
            return LetWhenAST(endtoken, tkn, False, vars_and_conds, stmts)

    # Visit a parse tree produced by HarmonyParser#opt_returns.
    def visitOpt_returns(self, ctx: HarmonyParser.Opt_returnsContext):
        return ctx.NAME()

    # Visit a parse tree produced by HarmonyParser#method_decl.
    def visitMethod_decl(self, ctx: HarmonyParser.Method_declContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        value = str(ctx.NAME())
        method_name = self.get_token(ctx.NAME().symbol, value)
        args = self.visit(ctx.bound()) if ctx.bound() else []
        if ctx.opt_returns() is None:
            result_name = None
        else:
            ors = self.visit(ctx.opt_returns())
            result_name = self.get_token(ors.symbol, str(ors))
        body = self.visit(ctx.block())
        colon = ctx.COLON()
        ctoken = self.get_token(colon.getSymbol(), colon.getText())
        ast = MethodAST(endtoken, tkn, False, method_name, args, result_name, body, ctoken)
        return ast

    # Visit a parse tree produced by HarmonyParser#while_block.
    def visitWhile_block(self, ctx: HarmonyParser.While_blockContext):
        cond = self.visit(ctx.expr())
        stmts = self.visit(ctx.block())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        colon = ctx.COLON()
        ctoken = self.get_token(colon.getSymbol(), colon.getText())
        return WhileAST(endtoken, tkn, False, cond, stmts, ctoken)

    # Visit a parse tree produced by HarmonyParser#if_block.
    def visitIf_block(self, ctx: HarmonyParser.If_blockContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        colon = ctx.COLON()
        ctoken = self.get_token(colon.getSymbol(), colon.getText())
        cond = self.visit(ctx.expr())
        if_block = self.visit(ctx.block())
        alts = [(cond, if_block, tkn, endtoken, ctoken)]
        else_block = None
        if ctx.elif_block() is not None:
            for c in ctx.elif_block():
                alts.append(self.visit(c))
        if ctx.else_block() is not None:
            else_block = self.visit(ctx.else_block())
        return IfAST(endtoken, tkn, False, alts, else_block)

    # Visit a parse tree produced by HarmonyParser#elif_block.
    def visitElif_block(self, ctx: HarmonyParser.Elif_blockContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        colon = ctx.COLON()
        ctoken = self.get_token(colon.getSymbol(), colon.getText())
        cond = self.visit(ctx.expr())
        stmts = self.visit(ctx.block())
        return cond, stmts, tkn, endtoken, ctoken

    # Visit a parse tree produced by HarmonyParser#else_block.
    def visitElse_block(self, ctx: HarmonyParser.Else_blockContext):
        return self.visit(ctx.block())

    # Visit a parse tree produced by HarmonyParser#var_stmt.
    def visitVar_stmt(self, ctx:HarmonyParser.Var_stmtContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        lv = self.visit(ctx.bound())
        e = self.visit(ctx.tuple_rule())
        return VarAST(endtoken, tkn, False, [(lv, e)])

    # Visit a parse tree produced by HarmonyParser#normal_block.
    def visitNormal_block(self, ctx: HarmonyParser.Normal_blockContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if ctx.block() is not None:
            return self.visit(ctx.block())
        elif ctx.block_stmts() is not None:
            stmts = self.visit(ctx.block_stmts())
            return BlockAST(endtoken, tkn, False, stmts, endtoken)
        raise HarmonyCompilerError(
            message="Unexpected block structure.",
            filename=self.file,
            line=tkn[2],
            column=tkn[3],
            lexeme=tkn[0]
        )

    # Visit a parse tree produced by HarmonyParser#block_stmts.
    def visitBlock_stmts(self, ctx: HarmonyParser.Block_stmtsContext):
        block = [s for stmt in ctx.stmt() for s in self.visit(stmt)]
        return [s for s in block if s is not None]

    # Visit a parse tree produced by HarmonyParser#label.
    def visitLabel(self, ctx: HarmonyParser.LabelContext):
        names = [self.get_token(n.symbol, str(n)) for n in ctx.NAME()]
        return names

    # Visit a parse tree produced by HarmonyParser#stmt.
    def visitStmt(self, ctx: HarmonyParser.StmtContext):
        labels = []
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if ctx.label() is not None:
            labels = self.visit(ctx.label())
        if ctx.one_line_stmt() is not None:
            stmt = self.visit(ctx.one_line_stmt())
        elif ctx.compound_stmt() is not None:
            stmt = self.visit(ctx.compound_stmt())
        elif ctx.import_stmt() is not None:
            stmt = self.visit(ctx.import_stmt())
        elif ctx.normal_block() is not None:
            stmt = self.visit(ctx.normal_block())
        else:
            return []

        assert stmt is not None
        if isinstance(stmt, AST):
            stmt_block = [LocationAST(endtoken, tkn, stmt, self.file, tkn[2])]
        elif isinstance(stmt, list):
            stmt_block = [LocationAST(endtoken, tkn, s, self.file, tkn[2]) for s in stmt]
        if ctx.COLON() or labels:
            if ctx.COLON():
                colon = ctx.COLON()
                ctoken = self.get_token(colon.getSymbol(), colon.getText())
                block = BlockAST(endtoken, tkn, False, stmt_block, ctoken)
            else:
                block = BlockAST(endtoken, tkn, False, stmt_block, endtoken)
            if labels:
                return [LocationAST(endtoken, tkn, LabelStatAST(endtoken, tkn, labels, block), self.file, tkn[2])]
            else:
                return [block]
        return stmt_block

    def visitBlock(self, ctx: HarmonyParser.BlockContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if ctx.normal_block():
            return self.visit(ctx.normal_block())
        elif ctx.one_line_stmt():
            return BlockAST(endtoken, tkn, False, self.visit(ctx.one_line_stmt()), endtoken)
        raise HarmonyCompilerError(
            message="Unable to parse block",
            filename=self.file,
            line=tkn[2],
            column=tkn[3],
            lexeme=tkn[0]
        )

    # Visit a parse tree produced by HarmonyParser#one_line_stmt.
    def visitOne_line_stmt(self, ctx: HarmonyParser.One_line_stmtContext):
        stmts = [self.visit(ctx.simple_stmt())]
        if ctx.one_line_stmt():
            stmts.extend(self.visit(ctx.one_line_stmt()))
        return stmts

    # Visit a parse tree produced by HarmonyParser#arith_op.
    def visitArith_op(self, ctx:HarmonyParser.Arith_opContext):
        return self.get_token(ctx.start, ctx.getText())

    # Visit a parse tree produced by HarmonyParser#set_rule_1.
    def visitSet_rule_1(self, ctx: HarmonyParser.Set_rule_1Context):
        if ctx.set_rule():
            return self.visit(ctx.set_rule())
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return SetAST(endtoken, self.get_token(ctx.start, ctx.start.text), [])

    # Visit a parse tree produced by HarmonyParser#empty_dict.
    def visitEmpty_dict(self, ctx: HarmonyParser.Empty_dictContext):
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return DictAST(endtoken, self.get_token(ctx.start, ctx.start.text), [])

    # Visit a parse tree produced by HarmonyParser#paren_tuple.
    def visitParen_tuple(self, ctx: HarmonyParser.Paren_tupleContext):
        if ctx.tuple_rule():
            return self.visit(ctx.tuple_rule())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return ConstantAST(endtoken, (emptytuple, self.file, tkn[2], tkn[3]))

    # Visit a parse tree produced by HarmonyParser#brack_tuple.
    def visitBracket_tuple(self, ctx: HarmonyParser.Bracket_tupleContext):
        if ctx.tuple_rule():
            return self.visit(ctx.tuple_rule())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return ConstantAST(endtoken, (emptytuple, self.file, tkn[2], tkn[3]))

    def visitLambda_expr(self, ctx: HarmonyParser.Lambda_exprContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        args = self.visit(ctx.bound()) if ctx.bound() is not None else []
        body = self.visit(ctx.nary_expr())
        return LambdaAST(endtoken, args, body, tkn, False)

    # Visit a parse tree produced by HarmonyParser#set_rule.
    def visitSet_rule(self, ctx: HarmonyParser.Set_ruleContext):
        values = [self.visit(e) for e in ctx.nary_expr()]
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if ctx.COLON():
            if ctx.iter_parse():
                k, v = values[0], values[1]
                comp = self.visit(ctx.iter_parse())
                return DictComprehensionAST(endtoken, k, v, comp, tkn)
            k = values[::2]
            v = values[1::2]
            return DictAST(endtoken, tkn, list(zip(k, v)))
        if ctx.iter_parse():
            comp = self.visit(ctx.iter_parse())
            return SetComprehensionAST(endtoken, values[0], comp, tkn)
        if ctx.RANGE():
            return RangeAST(endtoken, values[0], values[1], tkn)
        if ctx.COMMA():
            return SetAST(endtoken, tkn, values)
        return SetAST(endtoken, tkn, [values[0]])

    # Visit a parse tree produced by HarmonyParser#iter_parse.
    def visitIter_parse(self, ctx: HarmonyParser.Iter_parseContext):
        v = [self.visit(e) for e in ctx.children]
        return [x for x in v if x is not None]

    # Visit a parse tree produced by HarmonyParser#for_parse.
    def visitFor_parse(self, ctx: HarmonyParser.For_parseContext):
        l1 = self.visit(ctx.bound(0))
        l2 = self.visit(ctx.bound(1)) if ctx.bound(1) else None
        expr = self.visit(ctx.nary_expr())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if l2 is not None:
            return "for", (l2, l1, expr, tkn, endtoken)
        else:
            return "for", (l1, None, expr, tkn, endtoken)

    # Visit a parse tree produced by HarmonyParser#where_parse.
    def visitWhere_parse(self, ctx: HarmonyParser.Where_parseContext):
        expr = self.visit(ctx.nary_expr())
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        return "where", (expr, tkn, endtoken)

    # Visit a parse tree produced by HarmonyParser#tuple_rule.
    def visitTuple_rule(self, ctx: HarmonyParser.Tuple_ruleContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        values = [self.visit(e) for e in ctx.nary_expr()]
        if ctx.iter_parse():
            iterators = self.visit(ctx.iter_parse())
            return ListComprehensionAST(endtoken, values[0], iterators, tkn)
        comma_count = len(ctx.COMMA())
        if len(values) == 1 and comma_count != 1:
            return values[0]
        return TupleAST(endtoken, values, tkn)

    # Visit a parse tree produced by HarmonyParser#nary_expr.
    def visitNary_expr(self, ctx: HarmonyParser.Nary_exprContext):
        expressions = [self.visit(e) for e in ctx.expr_rule()]
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if ctx.comp_op():
            comps = [self.visit(o) for o in ctx.comp_op()]
            return CmpAST(endtoken, tkn, comps, expressions)
        if ctx.arith_op():
            ops = [self.visit(o) for o in ctx.arith_op()]
            expected = ops[0]
            if expected[0] not in self.associative_operators and len(ops) > 1:
                second_op = ops[1]
                raise HarmonyCompilerError(
                    message=f"Cannot have expression with multiple non-associative operators.",
                    filename=self.file,
                    line=second_op[2],
                    column=second_op[3],
                    lexeme=second_op[0]
                )
            for actual in ops[1:]:
                if actual[0] != expected[0]:
                    raise HarmonyCompilerError(
                        message=f"Cannot have multiple types of operators in one expression. Expected {expected[0]} but got {actual[0]}",
                        filename=self.file,
                        line=actual[2],
                        column=actual[3],
                        lexeme=actual[0]
                    )
            binop = ops[0]
            return NaryAST(endtoken, tkn, binop, expressions)
        if ctx.IF():
            condition = self.visit(ctx.nary_expr())
            expressions.insert(1, condition)
            op_token = self.get_token(ctx.IF().symbol, str(ctx.IF()))
            return NaryAST(endtoken, tkn, op_token, expressions)
        if ctx.IN():
            in_token = self.get_token(ctx.IN().symbol, str(ctx.IN()))
            ast = NaryAST(endtoken, tkn, in_token, expressions)
            if ctx.NOT():
                not_token = self.get_token(ctx.NOT().symbol, str(ctx.NOT()))
                return NaryAST(endtoken, tkn, not_token, [ast])
            return ast
        if ctx.IMPLIES():
            implies_token = self.get_token(ctx.IMPLIES().symbol, str(ctx.IMPLIES()))
            ast = NaryAST(endtoken, tkn, implies_token, expressions)
            if ctx.NOT():
                not_token = self.get_token(ctx.NOT().symbol, str(ctx.NOT()))
                return NaryAST(endtoken, tkn, not_token, [ast])
            return ast
        return expressions[0]

    # Visit a parse tree produced by HarmonyParser#expr_rule.
    def visitExpr_rule(self, ctx:HarmonyParser.Expr_ruleContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if ctx.SETINTLEVEL():
            expr = self.visit(ctx.expr_rule())
            return SetIntLevelAST(endtoken, tkn, expr)
        if ctx.SAVE():
            expr = self.visit(ctx.expr_rule())
            return SaveAST(endtoken, tkn, expr)
        if ctx.STOP():
            expr = self.visit(ctx.expr_rule())
            return StopAST(endtoken, tkn, expr)
        if ctx.unary_op():
            op = self.visit(ctx.unary_op())
            if op[0] == '!':
                expr = self.visit(ctx.expr_rule())
                return PointerAST(endtoken, expr, tkn)
            elif op[0] == '?':
                expr = self.visit(ctx.expr_rule())
                return AddressAST(endtoken, expr, tkn)
            else:
                expr = self.visit(ctx.expr_rule())
                return NaryAST(endtoken, tkn, op, [expr])
        if ctx.application():
            return self.visit(ctx.application())
        raise HarmonyCompilerError(
            message="Invalid expression",
            filename=self.file,
            line=tkn[2],
            column=tkn[3],
            lexeme=tkn[0]
        )

    # Visit a parse tree produced by HarmonyParser#application.
    def visitApplication(self, ctx:HarmonyParser.ApplicationContext):
        tkn = self.get_token(ctx.start, ctx.start.text)
        endtoken = self.get_token(ctx.stop, ctx.stop.text)
        if ctx.ARROWID():
            p = self.visit(ctx.application())
            name = self.get_token(ctx.ARROWID().symbol, str(ctx.ARROWID()))
            (lexeme, file, line, col) = name
            lexeme = lexeme[2:]
            col += 2
            while lexeme[0] == ' ':
                lexeme = lexeme[1:]
                col += 1
            return ApplyAST(endtoken, PointerAST(endtoken, p, tkn), ConstantAST(endtoken, (lexeme, file, line, col)), tkn)
        elif ctx.application():
            f = self.visit(ctx.application())
            return ApplyAST(endtoken, f, self.visit(ctx.basic_expr()), tkn)
        else:
            return self.visit(ctx.basic_expr())



import subprocess
import sys
import json
from typing import Any, Dict, List, Optional, Set, Tuple
from threading import Thread

from harmony_model_checker.harmony.jsonstring import json_string
from harmony_model_checker.iface import Transitions_t

try:
    import pydot  # type: ignore
    got_pydot = True
except Exception as e:
    got_pydot = False

try:
    from automata.fa.dfa import DFA  # type: ignore
    got_automata = True
except Exception as e:
    got_automata = False


# an error state is a non-final state all of whose outgoing transitions
# point only to itself
def find_error_states(transitions, final_states):
    error_states = set()
    for s, d in transitions.items():
        if s not in final_states and all(v == s for v in d.values()):
            error_states.add(s)
    return error_states

def is_dfa_equivalent(dfa: DFA, hfa: DFA) -> bool:
    stack: List[Tuple[int, int]] = []

    # Create states, where each state is renamed to an index to make each state unique
    dfa_states = [(k, v) for k, v in enumerate(list(dfa.states))]
    hfa_states = [(len(dfa_states) + k, v) for k, v in enumerate(list(hfa.states))]
    states = dfa_states + hfa_states
    n = len(states)

    # For convenience for mapping between states and indices
    idx_to_states = {k: v for k, v in states}
    dfa_state_to_idx = {v: k for k, v in dfa_states}
    hfa_state_to_idx = {v: k for k, v in hfa_states}

    dfa_final_states = {k: v for k, v in dfa_states if v in dfa.final_states}
    hfa_final_states = {k: v for k, v in hfa_states if v in hfa.final_states}
    # Collection of final states, indexed by state idx
    final_states = {**dfa_final_states, **hfa_final_states}

    # Tree-implementation of sets of states
    parents: List[int] = [0] * n
    rank: List[int] = [0] * n
    def make_set(q: int):
        parents[q] = q
        rank[q] = 0

    # Create a set for each state in the 2 DFAs
    for k, _ in states:
        make_set(k)

    def find_set(x):
        if x != parents[x]:
            parents[x] = find_set(parents[x])
        return parents[x]

    def link(x: int, y: int):
        if rank[x] > rank[y]:
            parents[y] = x
        else:
            parents[x] = y
            if rank[x] == rank[y]:
                rank[y] += 1

    def union(p: int, q: int):
        link(find_set(p), find_set(q))

    # Union the two initial states
    dfa_init_idx = dfa_state_to_idx[dfa.initial_state]
    hfa_init_idx = hfa_state_to_idx[hfa.initial_state]
    union(dfa_init_idx, hfa_init_idx)
    stack.append((dfa_init_idx, hfa_init_idx))

    # Traverse the stack
    while stack:
        k1, k2 = stack.pop()
        q1, q2 = idx_to_states[k1], idx_to_states[k2]
        dfa_t = dfa.transitions[q1]
        hfa_t = hfa.transitions[q2]

        # Check each common transitions
        symbols = set(dfa_t.keys()).intersection(hfa_t.keys())
        for s in symbols:
            p1 = dfa_state_to_idx[dfa_t[s]]
            p2 = hfa_state_to_idx[hfa_t[s]]

            r1 = find_set(p1)
            r2 = find_set(p2)
            # If their sets are not equivalent, then combine them
            if r1 != r2:
                union(p1, p2)
                stack.append((p1, p2))

    # Create sets of sets of states
    sets: Dict[int, set] = dict()
    for k, p in enumerate(parents):
        if p in sets:
            sets[p].add(k)
        else:
            sets[p] = {k}

    # Check condition for DFA equivalence
    return all(
        all(q in final_states for q in s) or all(q not in final_states for q in s)
        for s in sets.values()
    )

def read_hfa_file(file):
    try:
        with open(file, encoding='utf-8') as fd:
            js = json.load(fd, strict=False)
            initial = js["initial"]
            states = { "{}" }
            final = set()
            symbols = set()
            for e in js["edges"]:
                symbol = e["sym"]
                symbols.add(json_string(js["symbols"][symbol]))
            transitions = { "{}": { s: "{}" for s in symbols } }
            for n in js["nodes"]:
                idx: str = n["idx"]
                states.add(idx)
                if n["type"] == "final":
                    final.add(idx)
                transitions[idx] = { s: "{}" for s in symbols }
            for e in js["edges"]:
                symbol = e["sym"]
                transitions[e["src"]][json_string(js["symbols"][symbol])] = e["dst"]
        return DFA(
            states=states,
            input_symbols=symbols,
            transitions=transitions,
            initial_state=initial,
            final_states=final
        )
    except IOError:
        return False

def compare_behaviors(file, dfa):
    # TODO.  The following code can use read_hfa_file() I believe
    with open(file, encoding='utf-8') as fd:
        js = json.load(fd, strict=False)
        initial = js["initial"]
        states = { "{}" }
        final = set()
        symbols = set()
        for e in js["edges"]:
            symbol = e["sym"]
            symbols.add(json_string(js["symbols"][symbol]))
        transitions = { "{}": { s: "{}" for s in symbols } }
        for n in js["nodes"]:
            idx: str = n["idx"]
            states.add(idx)
            if n["type"] == "final":
                final.add(idx)
            transitions[idx] = { s: "{}" for s in symbols }
        for e in js["edges"]:
            symbol = e["sym"]
            transitions[e["src"]][json_string(js["symbols"][symbol])] = e["dst"]

    hfa = DFA(
        states=states,
        input_symbols=symbols,
        transitions=transitions,
        initial_state=initial,
        final_states=final
    )

    print("Phase 7: comparing behaviors", len(dfa.states), len(hfa.states))
    if len(dfa.states) == len(hfa.states):      # HACK
        return                                  # HACK
    if len(dfa.states) > 100 or len(hfa.states) > 100:
        print("  warning: this could take a while")

    assert dfa.input_symbols <= hfa.input_symbols
    if dfa.input_symbols < hfa.input_symbols:
        print("behavior warning: symbols missing from behavior:",
            hfa.input_symbols - dfa.input_symbols)
        return

    if not is_dfa_equivalent(dfa, hfa):
        print("behavior warning: subset of specified behavior")
        diff = hfa - dfa
        behavior_show_diagram(diff, "diff.png")

# Modified from automata-lib
def behavior_show_diagram(dfa: DFA, path=None):
    graph = pydot.Dot(graph_type='digraph', rankdir='LR')
    nodes = {}
    rename: Dict[str, int] = {}
    next_idx = 0
    error_states = find_error_states(dfa.transitions, dfa.final_states)
    for state in dfa.states:
        if state in rename:
            idx = rename[state]
        else:
            rename[state] = idx = next_idx
            next_idx += 1
        if state in error_states:
            continue
        if state == dfa.initial_state:
            # color start state with green
            if state in dfa.final_states:
                initial_state_node = pydot.Node(
                    str(idx),
                    style='filled',
                    peripheries=2,
                    fillcolor='#66cc33', label="initial")
            else:
                initial_state_node = pydot.Node(
                    str(idx), style='filled', fillcolor='#66cc33', label="initial")
            nodes[state] = initial_state_node
            graph.add_node(initial_state_node)
        else:
            if state in dfa.final_states:
                state_node = pydot.Node(str(idx), peripheries=2, label="final")
            else:
                state_node = pydot.Node(str(idx), label="")
            nodes[state] = state_node
            graph.add_node(state_node)
    # adding edges
    for from_state, lookup in dfa.transitions.items():
        for to_label, to_state in lookup.items():
            if to_state not in error_states and to_label != "":
                graph.add_edge(pydot.Edge(
                    nodes[from_state],
                    nodes[to_state],
                    label=to_label
                ))
    if path:
        try:
            graph.write_png(path)
        except FileNotFoundError:
            print("install graphviz (www.graphviz.org) to see output DFAs")
    return graph

def eps_closure_rec(states: Set[str], transitions: Transitions_t, current: str, output: Set[str]):
    if current in output:
        return
    output.add(current)
    t = transitions[current]
    if '' in t:
        for s in t['']:
            eps_closure_rec(states, transitions, s, output)

def eps_closure(states: Set[str], transitions: Transitions_t, current: str):
    x: Set[str] = set()
    eps_closure_rec(states, transitions, current, x)
    return frozenset(x)

minified_dfa = None

def do_minify(intermediate_dfa):
    global minified_dfa
    minified_dfa = intermediate_dfa.minify(retain_names = True)

def behavior_parse(js, minify, outputfiles, behavior):
    if outputfiles["hfa"] is None and outputfiles["png"] is None and outputfiles["gv"] is None and behavior is None:
        return

    if "dfasize" in js and js["dfasize"] > 1024:
        print("    * too many states for post-analysis") 
        if outputfiles["png"] is not None:
            print("    * png output file suppressed") 
        elif outputfiles["gv"] is not None:
            print("    * gv output file suppressed") 
        return

    minify = outputfiles["png"] is not None or outputfiles["gv"] is not None

    # Read the hfa file
    # TODO: only do this if there are not too many nodes
    if outputfiles["hfa"] is None:
        return
    dfa = read_hfa_file(outputfiles["hfa"])
    if not dfa:
        return

    if got_automata and minify:
        intermediate_dfa = dfa
        if True or len(intermediate_dfa.states) > 100: 
            print("    * minify #states=%d"%len(intermediate_dfa.states))
        thread = Thread(target=do_minify, args=(intermediate_dfa,))
        thread.start()
        thread.join(15)
        if thread.is_alive():
            print("    * minify: taking too long and giving up (not fatal)")
        else:
            global minified_dfa
            dfa = minified_dfa
            if True or len(intermediate_dfa.states) > 100: 
                print("    * minify done #states=%d"%len(dfa.states))
    dfa_states = dfa.states
    (dfa_transitions,) = dfa.transitions,
    dfa_initial_state = dfa.initial_state
    dfa_final_states = dfa.final_states
    dfa_error_states = set()

    if outputfiles["gv"] is not None:
        with open(outputfiles["gv"], "w", encoding='utf-8') as fd:
            names = {}
            for (i, s) in enumerate(dfa_states):
                names[s] = i
            print("digraph {", file=fd)
            print("  rankdir = \"LR\"", file=fd)
            for s in dfa_states:
                if s in dfa_error_states:
                    continue
                if s == dfa_initial_state:
                    if s in dfa_final_states:
                        print("  s%s [label=\"initial\",style=filled,peripheries=2,fillcolor=\"#66cc33\"]"%names[s], file=fd)
                    else:
                        print("  s%s [label=\"initial\",style=filled,fillcolor=\"#66cc33\"]"%names[s], file=fd)
                else:
                    if s in dfa_final_states:
                        print("  s%s [peripheries=2,label=\"final\"]"%names[s], file=fd)
                    else:
                        print("  s%s [label=\"\"]"%names[s], file=fd)

            for (src, edges) in dfa_transitions.items():
                for (input, dst) in edges.items():
                    if dst not in dfa_error_states:
                        print("  s%s -> s%s [label=%s]"%(names[src], names[dst], json.dumps(input, ensure_ascii=False)), file=fd)
            print("}", file=fd)

    if outputfiles["png"] is not None:
        if len(dfa.states) > 250:
            print("    * output of png file suppressed (too many states)") 
        else:
            if got_pydot and got_automata:
                behavior_show_diagram(dfa, path=outputfiles["png"])
            else:
                print("    * running dot", len(dfa.states))
                assert outputfiles["gv"] is not None
                try:
                    subprocess.run(["dot", "-Tpng", "-o", outputfiles["png"],
                                    outputfiles["gv"] ])
                except FileNotFoundError:
                    print("install graphviz (www.graphviz.org) to see output DFAs")

    if behavior is not None:
        if got_automata:
            if "suppressed" in js:
                print("behavior subset checking suppressed; use -b flag to enable")
            else:
                compare_behaviors(behavior, dfa)
        else:
            print("Can't check behavior subset because automata-lib is not available")



import sys
import getopt
import json
from typing import Dict, Optional, Set
from automata.fa.nfa import NFA # type: ignore
from automata.fa.dfa import DFA # type: ignore

Transitions_t = Dict[str, Dict[str, Set[str]]]

def dfadump(dfa):
    rename = {}
    for i, s in enumerate(dfa.states):
        rename[s] = "{}" if s == "{}" else str(i)
    print("states", { rename[s] for s in dfa.states }, file=sys.stderr)
    for (s, d) in dfa.transitions.items():
        print(rename[s], ":", file=sys.stderr)
        for t, s2 in d.items():
            print("   ", t, ":", rename[s2], file=sys.stderr)

# Get rid of choose states
def dechoose(states, transitions, choose_states):
    print("dechoose", file=sys.stderr)
    for (k, v) in states.items():
        if k in choose_states:
            # See what incoming nodes k has
            incoming = set()
            for k2 in states:
                if v in transitions[k2] and k in transitions[k2][v]:
                    incoming.add(k2)
            for k2 in incoming:
                # remove the existing transition from the incoming node
                transitions[k2][v].remove(k)
                if transitions[k2][v] == set():
                    del transitions[k2][v]

                # distribute each outgoing edge among the incoming node
                for (v2, s) in transitions[k].items():
                    for out in s:
                        if k2 != out:
                            if v2 in transitions[k2]:
                                transitions[k2][v2].add(out)
                            else:
                                transitions[k2][v2] = { out }

    for k in choose_states:
        del states[k]
        del transitions[k]

# Get rid of stutter transitions
def destutter(states: Dict[str, str], transitions: Transitions_t):
    print("destutter", len(states), file=sys.stderr)

    # figure out for each state what other states point to it
    pointsto: Dict[str, Set[str]] = { k:set() for k in states }
    for src in states:
        for (v, s) in transitions[src].items():
            for dst in s:
                pointsto[dst].add(src)

    updated = set()
    for (k, v) in states.items():
        # See if state k has an outgoing edge with its value
        if v in transitions[k]:
            # See what incoming nodes k has
            incoming = set()
            for k2 in pointsto[k]:
                if v in transitions[k2]:
                    assert k in transitions[k2][v]
                    incoming.add(k2)
            # distribute each outgoing edge among the incoming nodes
            for out in transitions[k][v]:
                for k2 in incoming:
                    if k2 != out:
                        transitions[k2][v].add(out)
                        pointsto[k].add(k2)

            # remove the outgoing edges
            del transitions[k][v]

            updated.add(k)

    if updated == set():
        return False

    # Remove states with no outgoing edges
    for src in states:
        for v in transitions[src]:
            toremove = set()
            for dst in transitions[src][v]:
                if transitions[dst] == {}:
                    toremove.add(dst)
            if toremove != set():
                transitions[src][v] -= toremove
    for k in updated:
        if transitions[k] == {}:
            del states[k]
            del transitions[k]

    return True

def parse(js, outfmt, minify):
    states: Dict[str, str] = {}
    initial_state: Optional[str] = None
    final_states: Set[str] = set()
    choose_states: Set[str] = set()
    input_symbols = { "__term__" }
    transitions: Transitions_t = {}

    for s in js["nodes"]:
        idx = str(s["idx"])
        val = str(s["value"])
        input_symbols.add(val)
        transitions[idx] = {}
        if s["type"] == "initial":
            assert initial_state is None
            initial_state = idx
            val = "__init__"
        elif s["type"] == "terminal":
            final_states.add(idx)
        elif s["type"] == "choose" and int(s["choosing_atomic_level"]) > 0:
            choose_states.add(idx)
        states[idx] = val

    for edge in js['edges']:
        src = str(edge["src"])
        dst = str(edge["dst"])
        assert dst != initial_state
        assert src not in final_states
        if src != dst:
            if dst in final_states:
                val = "__term__"
            else:
                val = states[dst]
            if val in transitions[src]:
                transitions[src][val].add(dst)
            else:
                transitions[src][val] = {dst}

    dechoose(states, transitions, choose_states)

    while destutter(states, transitions):
        pass

    # print("states", states, file=sys.stderr)
    # print("final", final_states, file=sys.stderr)
    # print("symbols", input_symbols, file=sys.stderr)
    # print("transitions", transitions, file=sys.stderr)

    nfa = NFA(
        states=set(states.keys()),
        input_symbols=input_symbols,
        transitions=transitions,
        initial_state=initial_state,
        final_states=final_states
    )

    intermediate = DFA.from_nfa(nfa)  # returns an equivalent DFA

    # TODO.  Minifying the DFA can lead to results where not all incoming
    #        edges to a node are labeled the same and other stuff.
    if minify:
        dfa = intermediate.minify()
    else:
        dfa = intermediate

    # dfadump(dfa)
    # sys.exit(0)

    # dfa.show_diagram(path='./dfa1.png')
    # sys.exit(0)

    # Give each state a simple integer name
    names = {}
    values = {}
    for (i, s) in enumerate(dfa.states):
        names[s] = i
        values[s] = "???"
    values[dfa.initial_state] = "initial"

    # Figure out the value of each of the states
    for (src, edges) in dfa.transitions.items():
        for (input, dst) in edges.items():
            if dst != '{}' and src != dst:
                values[dst] = input

    if outfmt == "dot":
        print("digraph {")

        for s in dfa.states:
            if s == "{}":
                continue
            if s == dfa.initial_state:
                print("  s%s [label=\"%s\",shape=octagon]"%(names[s], values[s]))
            elif s in dfa.final_states:
                print("  s%s [label=\"%s\",shape=doubleoctagon]"%(names[s], values[s]))
            else:
                print("  s%s [label=\"%s\",shape=box]"%(names[s], values[s]))

        for (src, edges) in dfa.transitions.items():
            for (input, dst) in edges.items():
                if dst != '{}' and src != dst:
                    print("  s%s -> s%s"%(names[src], names[dst]))

        print("}")
    else:       # json format, same as input
        assert outfmt == "json"
        print("{")
        print("  \"nodes\": [")

        first = True
        for s in dfa.states:
            if s == "{}":
                continue
            if first:
                first = False
            else:
                print(",")
            print("    {")
            print("      \"idx\": \"%s\","%names[s])
            print("      \"value\": \"%s\","%values[s])
            if s == dfa.initial_state:
                t = "initial"
            elif s in dfa.final_states:
                t = "final"
            else:
                t = "normal"
            print("      \"type\": \"%s\""%t)
            print("    }", end="")
        print()
        print("  ],")

        print("  \"edges\": [")
        first = True
        for (src, edges) in dfa.transitions.items():
            for (input, dst) in edges.items():
                if dst != '{}' and src != dst:
                    if first:
                        first = False
                    else:
                        print(",")
                    print("    {")
                    print("      \"src\": \"%s\","%names[src])
                    print("      \"dst\": \"%s\""%names[dst])
                    print("    }", end="")
        print()
        print("  ]")

        print("}")

def usage():
    print("Usage: iface [-T type] [-M] file.json", file=sys.stderr)
    sys.exit(1)

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "MT:",
                ["type=", "minify", "help"])
    except getopt.GetoptError as err:
        print(str(err))
        usage()
    outfmt = "dot"
    minify = False
    for o, a in opts:
        if o in [ "-T", "--type"]:
            if a not in [ "dot", "json" ]:
                print("type must be dot or json", file=sys.stderr)
                sys.exit(1)
            outfmt = a
        elif o in [ "-M", "--minify" ]:
            minify = True
        else:
            usage()

    if args == []:
        usage()

    file = args[0]
    with open(file, encoding='utf-8') as f:
        js = json.load(f, strict=False)
        parse(js, outfmt, minify);

if __name__ == "__main__":
    main()



import sys
from harmony_model_checker.harmony.value import emptydict, DictValue


class State:

    def __init__(self, code, labels):
        self.code = code
        self.labels = labels
        self.vars = emptydict
        self.ctxbag = {}        # running contexts
        self.stopbag = {}       # stopped contexts
        self.termbag = {}       # terminated contexts
        self.choosing = None
        self.invariants = set()
        self.initializing = True

    def __repr__(self):
        return "State(" + str(self.vars) + ", " + str(self.ctxbag) + ", " + \
               str(self.stopbag) + ", " + str(self.invariants) + ")"

    def __hash__(self):
        h = self.vars.__hash__()
        for c in self.ctxbag.items():
            h ^= c.__hash__()
        for c in self.stopbag.items():
            h ^= c.__hash__()
        for c in self.termbag.items():
            h ^= c.__hash__()
        for i in self.invariants:
            h ^= i
        return h

    def __eq__(self, other):
        if not isinstance(other, State):
            return False
        assert self.code == other.code and self.labels == other.labels
        if self.vars != other.vars:
            return False
        if self.ctxbag != other.ctxbag:
            return False
        if self.stopbag != other.stopbag:
            return False
        if self.termbag != other.termbag:
            return False
        if self.choosing != other.choosing:
            return False
        if self.invariants != other.invariants:
            return False
        if self.initializing != self.initializing:
            return False
        return True

    def copy(self):
        s = State(self.code, self.labels)
        s.vars = self.vars      # no need to copy as store operations do it
        s.ctxbag = self.ctxbag.copy()
        s.stopbag = self.stopbag.copy()
        s.termbag = self.termbag.copy()
        s.choosing = self.choosing
        s.invariants = self.invariants.copy()
        s.initializing = self.initializing
        return s

    def get(self, var):
        return self.vars.d[var]

    def iget(self, indexes):
        path = indexes
        v = self.vars
        while indexes != []:
            try:
                v = v.d[indexes[0]]
            except KeyError:
                print()
                print("no index", indexes[0], "in variable", path)
                sys.exit(1)
            indexes = indexes[1:]
        return v

    def update(self, record, indexes, val):
        if len(indexes) > 1:
            v = self.update(record.d[indexes[0]], indexes[1:], val)
        else:
            v = val
        d = record.d.copy()
        d[indexes[0]] = v
        return DictValue(d)

    def doDelete(self, record, indexes):
        d = record.d.copy()
        if len(indexes) > 1:
            d[indexes[0]] = self.doDelete(record.d[indexes[0]], indexes[1:])
        else:
            del d[indexes[0]]
        return DictValue(d)

    def doStop(self, record, indexes, ctx):
        d = record.d.copy()
        if len(indexes) > 1:
            d[indexes[0]] = self.doStop(record.d[indexes[0]], indexes[1:], ctx)
        else:
            # TODO.  Should be print + set failure
            list = d[indexes[0]]
            assert(isinstance(list, DictValue))
            d2 = list.d.copy()
            d2[len(d2)] = ctx
            d[indexes[0]] = DictValue(d2)
        return DictValue(d)

    def set(self, indexes, val):
        self.vars = self.update(self.vars, indexes, val)

    def delete(self, indexes):
        self.vars = self.doDelete(self.vars, indexes)

    def stop(self, indexes, ctx):
        self.vars = self.doStop(self.vars, indexes, ctx)
        cnt = self.stopbag.get(ctx)
        if cnt is None:
            self.stopbag[ctx] = 1
        else:
            self.stopbag[ctx] = cnt + 1




import json
import copy
from harmony_model_checker.harmony.verbose import verbose_string

"""
Takes in filenames. Returns summary in string.
"""
def summaryMain(filenames, hco):
    # hco = {}
    hvm = {}
    summaryMain.microSteps = []
    # summaryMain.stackTraceList is a list of string with length <number of threads>. Each entry is the stack trace for each thread
    summaryMain.stackTraceList = []
    # summaryMain.stackTraceTextList contains stack trace to display at each microstep
    summaryMain.stackTraceTextList = []
    # summaryMain.threadMode is a list of thread modes at each microstep
    summaryMain.threadMode = []
    summaryMain.threadNumber = -1
    summaryMain.threadNames = []

    def constructMicrosteps():
        summaryMain.microSteps = []
        # initialize summaryMain.threadNumber to be number of threads
        summaryMain.threadNumber = len(hco['macrosteps'][-1]['contexts']) # fix bug - change 0 to -1
        # initialize summaryMain.microSteps
        for macroStep in hco['macrosteps']:
            for microStep in macroStep['microsteps']:
                cpMicroStep = dict(microStep)
                cpMicroStep['tid'] = macroStep['tid']
                cpMicroStep['name'] = macroStep['name']
                # cpMicroStep['invfails'] = macroStep['invfails']
                cpMicroStep['contexts'] = macroStep['contexts']
                cpMicroStep['context'] = macroStep['context']
                summaryMain.microSteps.append(cpMicroStep)
        # initialize summaryMain.stackTraceList
        summaryMain.stackTraceList = []
        summaryMain.stackTraceTextList = []
        for i in range(summaryMain.threadNumber):
            summaryMain.stackTraceList.append("")
        # initialize 
        for i in range(len(summaryMain.microSteps)):
            summaryMain.stackTraceTextList.append("")
            
        
    def constructStackTraceTextList():
        assert len(summaryMain.stackTraceList) == summaryMain.threadNumber
        assert len(summaryMain.stackTraceTextList) == len(summaryMain.microSteps)
        # base case for i = 0 to fix the initial stack trace
        assert 'contexts' in hco['macrosteps'][0]
        for thread in hco['macrosteps'][0]['contexts']:
            tid = int(thread['tid'])
            assert 'trace' in thread
            trace = thread['trace']
            traceLine = ""
            assert len(trace) > 0
            for j in range(len(trace) - 1):
                traceLine = traceLine + trace[j]['method'] + ' -> '
            traceLine += trace[len(trace) - 1]['method']
            summaryMain.stackTraceList[tid] = traceLine
        summaryMain.stackTraceTextList[0] = []
        for stackTraceLine in summaryMain.stackTraceList:
            summaryMain.stackTraceTextList[0].append(stackTraceLine)
        # other microsteps for i > 0
        for i in range(1, len(summaryMain.microSteps)):
            if 'trace' in summaryMain.microSteps[i]:
                # there is a change in stack trace
                trace = summaryMain.microSteps[i]['trace']
                traceLine = ""
                if(len(trace) > 0):
                    for j in range(len(trace) - 1):
                        traceLine = traceLine + trace[j]['method'] + ' -> '
                    traceLine = traceLine + trace[len(trace) - 1]['method']
                tid = int(summaryMain.microSteps[i]['tid'])
                summaryMain.stackTraceList[tid] = traceLine
                # update the ith entry in summaryMain.stackTraceTextList
                summaryMain.stackTraceTextList[i] = []
                for stackTraceLine in summaryMain.stackTraceList:
                    summaryMain.stackTraceTextList[i].append(stackTraceLine)
            else:
                # there is no change in stack trace
                summaryMain.stackTraceTextList[i] = copy.deepcopy(summaryMain.stackTraceTextList[i - 1])
            # print failure in stack trace (commented out since already covered by next)
            # if 'failure' in summaryMain.microSteps[i]:
            #     tid = int(summaryMain.microSteps[i]['tid'])
            #     summaryMain.stackTraceTextList[i][tid] += f" -> {summaryMain.microSteps[i]['failure']}!"
            
        # add "about to" information to the end of stack trace line
        i = 0
        for macrostep in hco['macrosteps']:
            i += len(macrostep['microsteps']) - 1
            assert 'contexts' in macrostep
            for context in macrostep['contexts']:
                if 'next' in context and int(context['tid']) == int(summaryMain.microSteps[i]['tid']):
                    tid = int(summaryMain.microSteps[i]['tid'])
                    summaryMain.stackTraceTextList[i][tid] += f" ({about(context)})"
                    tmp = i + 1
                    while tmp < len(summaryMain.microSteps) and int(summaryMain.microSteps[tmp]['tid']) != tid:
                        summaryMain.stackTraceTextList[tmp][tid] += f" ({about(context)})"
                        tmp += 1
            i += 1

        for i in range(len(summaryMain.stackTraceTextList)):
            for j in range(len(summaryMain.stackTraceTextList[i])):
                summaryMain.stackTraceTextList[i][j] = f"[{summaryMain.threadMode[i][j]}] " + summaryMain.stackTraceTextList[i][j] 
        
        # construct thread names
        for t_name in summaryMain.stackTraceTextList[-1]:
            summaryMain.threadNames.append(getThreadName(t_name))

    def constructStackTopDisplay():
        """
        construct stackTopDisplay
        """
        stacks = []
        stackTopDisplay = []
        for i in range(summaryMain.threadNumber):
            stacks.append([])
        # default stack for each thread
        for t in range(summaryMain.threadNumber):
            i = 0
            while i < len(summaryMain.microSteps) and int(summaryMain.microSteps[i]['tid']) != t:
                i += 1
            # fix bug for threads that don't run
            if i == len(summaryMain.microSteps):
                continue
            stacks[t] = summaryMain.microSteps[i]['context']['stack']
        framePointer = -1
        for i in range(len(summaryMain.microSteps)):
            tid = int(summaryMain.microSteps[i]['tid'])
            # update frame pointer
            if 'fp' in summaryMain.microSteps[i]:
                framePointer = int(summaryMain.microSteps[i]['fp'])
            # pop items first
            if 'pop' in summaryMain.microSteps[i]:
                assert len(stacks[tid]) >= int(summaryMain.microSteps[i]['pop'])
                for j in range(int(summaryMain.microSteps[i]['pop'])):
                    del stacks[tid][-1]
            if 'push' in summaryMain.microSteps[i]:
                for variable in summaryMain.microSteps[i]['push']:
                    stacks[tid].append(variable)
            # update stackTopDisplay
            stackTopDisplay.append([])
            for k in range(framePointer, len(stacks[tid])):
                stackTopDisplay[i].append(stacks[tid][k])

    def constructThreadMode():
        # initialize summaryMain.threadMode
        for i in range(len(summaryMain.microSteps)):
            summaryMain.threadMode.append([])
            for j in range(summaryMain.threadNumber):
                summaryMain.threadMode[i].append("runnable")
        # construct thread mode for each microstep
        i = 0
        for macrostep in hco['macrosteps']:
            if i > 0:
                summaryMain.threadMode[i] = copy.deepcopy(summaryMain.threadMode[i - 1])
            if 'contexts' in macrostep:
                for context in macrostep['contexts']:
                    if 'mode' in context:
                        summaryMain.threadMode[i][int(context['tid'])] = 'runnable' if context['mode'] == 'choosing' else context['mode'] # change choosing to runnable
            if ('context' in macrostep) and ('mode' in macrostep['context']):
                summaryMain.threadMode[i][int(macrostep['context']['tid'])] = 'runnable' if macrostep['context']['mode'] == 'choosing' else macrostep['context']['mode']
            for j in range(len(macrostep['microsteps'])):
                if j > 0:
                    summaryMain.threadMode[i] = copy.deepcopy(summaryMain.threadMode[i - 1])
                if 'mode' in macrostep['microsteps'][j]:
                    summaryMain.threadMode[i][int(summaryMain.microSteps[i]['tid'])] = 'runnable' if macrostep['microsteps'][j]['mode'] == 'choosing' else macrostep['microsteps'][j]['mode']
                if j < len(macrostep['microsteps']) - 1:
                    i += 1
            if 'contexts' in macrostep:
                for context in macrostep['contexts']:
                    if 'mode' in context:
                        summaryMain.threadMode[i][int(context['tid'])] = 'runnable' if context['mode'] == 'choosing' else context['mode']
            i += 1
        # for k in range(len(summaryMain.threadMode)):
        #     print(summaryMain.threadMode[k])
        

    def friendlyOutput():
        issueText = hco["issue"]
        # output = f"{issueText}\n"
        output = ""
        if issueText == "No issues":
            output += "Congratulations, your program has no issues!\n"
            return
        elif issueText == "Safety violation":
            output += "Unfortunately, there exists some state in your program that violates a safety property. \n"
            output += "Here is an execution trace summary that exhibits the issue:\n"
        elif issueText == "Invariant violation":
            output += "Unfortunately, an invariant in your program can violate. \n"
            output += "Here is an execution trace summary that exhibits the issue:\n"
        elif issueText == "Finally predicate violation":
            output += "Unfortunately, the finally assertion in your program can fail. \n"
            output += "Here is an execution trace summary that exhibits the issue:\n"
        elif issueText == "Behavior Violation: terminal state not final":
            pass
        elif issueText == "Non-terminating state":
            output += "Unfortunately, there exists some case where your program can't terminate. \n"
            output += "Here is an execution trace summary that exhibits the issue:\n"
        elif issueText == "Active busy waiting":
            output += "Unfortunately, there is active busy waiting in your program. That is, a thread is waiting for a condition while changing the state. \n"
            output += "Here is an execution trace summary that exhibits the issue:\n"
        elif "Data race" in issueText:
            output += f"Unfortunately, there is a data race in your program on {getDataRaceTarget(issueText)}. \n"
            output += "Here is an execution trace summary that exhibits the issue:\n"
        else:
            print(issueText)
            raise Exception("issue match failure")
        output += friendlyTrace()
        output += "\n"
        output += "More detailed information can be found in htm output or HarmonyGUI.\n"
        return output
    

    def friendlyTrace():
        trace = ""
        counter = 0
        subCounter = 0
        # trace += f"{counter}. Program starts with the initial thread T{int(summaryMain.microSteps[0]['tid'])} {summaryMain.threadNames[0]} at line {getStartLine(0)}.\n"
        trace += f"{counter}. Program starts with the initial thread T{int(summaryMain.microSteps[0]['tid'])}\n"
        for i in range(len(summaryMain.microSteps)):
            if (i == len(summaryMain.microSteps) - 1):
                break
            if (summaryMain.microSteps[i]['tid'] != summaryMain.microSteps[i + 1]['tid']):
                prev_tid = int(summaryMain.microSteps[i]['tid'])
                prev_tname = summaryMain.threadNames[prev_tid]
                next_tid = int(summaryMain.microSteps[i + 1]['tid'])
                next_tname = summaryMain.threadNames[next_tid]
                i_line_prev = i
                while (getTid(i) != getTid(i_line_prev) or getModule(i_line_prev) != "__main__"):
                    i_line_prev -= 1

                i_line_next = i + 1
                while (getTid(i + 1) != getTid(i_line_next) or getModule(i_line_next) != "__main__"):
                    i_line_next -= 1
                counter += 1
                subCounter = 0
                # last thread just terminated
                if getStatus(summaryMain.stackTraceTextList[i][prev_tid]) == "terminated":
                    trace += f"{counter}. At line {getEndLine(i_line_prev)}, T{prev_tid} {prev_tname} terminates. Context switch to T{next_tid} {next_tname} at line {getStartLine(i_line_next)}.\n"
                # last thread is about to do sth
                elif getAboutTo(summaryMain.stackTraceTextList[i][prev_tid]) != "":
                    aboutTo = getAboutTo(summaryMain.stackTraceTextList[i][prev_tid])
                    trace += f"{counter}. At line {getEndLine(i_line_prev)}, T{prev_tid} {prev_tname} is {aboutTo}. Context switch to T{next_tid} {next_tname} at line {getStartLine(i_line_next)}.\n"
                # nothing special
                else:
                    trace += f"{counter}. At line {getEndLine(i_line_prev)}, T{prev_tid} {prev_tname} is preempted. Context switch to T{next_tid} {next_tname} at line {getStartLine(i_line_next)}.\n"
        
            if i < len(summaryMain.microSteps):
                sharedVariableListBefore = getSharedVariable(i)
                sharedVariableListAfter =  getSharedVariable(i + 1)

                prev_tid = int(summaryMain.microSteps[i]['tid'])
                prev_tname = summaryMain.threadNames[prev_tid]
                i_line_prev = i
                while (getTid(i) != getTid(i_line_prev) or getModule(i_line_prev) != "__main__"):
                    i_line_prev -= 1
                    
                if sharedVariableListBefore != sharedVariableListAfter:
                    for variableName, variable in sharedVariableListAfter.items():
                        if (sharedVariableListBefore == None) or (variableName not in sharedVariableListBefore):
                            subCounter += 1
                            trace += f"    At line {getEndLine(i_line_prev)}, T{prev_tid} {prev_tname} initializes global variable <{variableName}> to {verbose_string(variable)}. \n"
                        elif sharedVariableListBefore[variableName] != sharedVariableListAfter[variableName]:
                            variableBefore = sharedVariableListBefore[variableName]
                            subCounter += 1
                            lst = getDifferentEntryTrace(variableBefore, variable)
                            if len(lst) > 0:
                                traceStr = ""
                                for e in lst:
                                    traceStr += f"[{e if isinstance(e, int) else verbose_string(e)}]"
                                trace += (f"    At line {getEndLine(i_line_prev)}, T{prev_tid} updates {variableName + traceStr} to {entryChangeTo(lst, variable)}. Global variable <{variableName}> becomes {verbose_string(variable)}. \n")
                            else: # len(lst) == 0
                                trace += (f"    At line {getEndLine(i_line_prev)}, T{prev_tid} updates global variable <{variableName}> to {verbose_string(variable)}. \n")
                        

        # a failure in the end
        if 'failure' in summaryMain.microSteps[len(summaryMain.microSteps) - 1]:
            counter += 1
            i_failure = len(summaryMain.microSteps) - 1
            prev_tid = getTid(i_failure)
            prev_tname = summaryMain.threadNames[prev_tid]
            failure_msg = summaryMain.microSteps[i_failure]['failure']
            trace += f"{counter}. At line {getEndLine(i_failure)} of T{prev_tid} {prev_tname}, {failure_msg}\n"
        return trace

    def getTid(i):
        return int(summaryMain.microSteps[i]['tid'])

    def getPc(i):
        return int(summaryMain.microSteps[i]["pc"])
    
    # return example: {'module': 'synch', 'endline': 34, 'stmt': [34, 1, 34, 21], 'line': 34, 'endcolumn': 21, 'column': 1}
    def getModule(i):
        return hvm['locs'][getPc(i)]['module']
    
    def getEndLine(i):
        loc = hvm['locs'][getPc(i)]
        if loc['module'] == "__main__":
            return loc['endline']
        return loc['module'] + ":" + str(loc['endline'])
    
    def getStartLine(i):
        loc = hvm['locs'][getPc(i)]
        if loc['module'] == "__main__":
            return loc['line']
        return loc['module'] + ":" + str(loc['line'])

    def getStatus(s):
        start = s.find("[")
        end = s.find("]")
        return s[start + 1:end]
    
    # example return: "" or "about to execute atomic section"
    def getAboutTo(s):
        start = s.find("(about to")
        if start < 0:
            return ""
        return s[(start + 1):-1]

    def getThreadName(s):
        start = s.find("]")
        end = s.find(")")
        return s[(start + 2):(end + 1)]
    
    def getDataRaceTarget(s):
        return s[11:-1]

    def getSharedVariable(microStepPointer):
        # clear all displayed content
        if microStepPointer == 0:
            return None
        microStepPointer -= 1
        # find current global variable state
        mostRecentSharedVariablePointer = microStepPointer
        while 'shared' not in summaryMain.microSteps[mostRecentSharedVariablePointer]:
            mostRecentSharedVariablePointer -= 1
            if mostRecentSharedVariablePointer < 0:
                return
            
        microstep = summaryMain.microSteps[mostRecentSharedVariablePointer]
        sharedVariableList = microstep['shared']
        return sharedVariableList

    def getDifferentEntryTrace(variableBefore, variableAfter):
        res = []
        if (variableBefore['type'] == variableAfter['type'] == 'list'): # list case
            # must have exactly one entry differ
            if (len(variableBefore['value']) == len(variableAfter['value'])) and len(variableBefore['value']) > 0:
                differCount = 0
                candidateAppend = None
                for i in range(len(variableBefore['value'])):
                    if variableBefore['value'][i] != variableAfter['value'][i]:
                        differCount += 1
                        candidateAppend = i
                if differCount == 1:
                    res.append(candidateAppend)
                    return (res + getDifferentEntryTrace(variableBefore['value'][candidateAppend], variableAfter['value'][candidateAppend]))
                
        elif (variableBefore['type'] == variableAfter['type'] == 'dict'): # dict case
            # must have exactly one entry differ
            if (len(variableBefore['value']) == len(variableAfter['value'])) and len(variableBefore['value']) > 0:
                differCount = 0
                candidateKey = None
                candidateValue = None
                matched = []
                for i in range(len(variableBefore['value'])):
                    if variableBefore['value'][i] not in variableAfter['value']:
                        differCount += 1
                        candidateKey = variableBefore['value'][i]['key']
                        candidateValue = variableBefore['value'][i]['value']
                    else:
                        matched.append(variableBefore['value'][i])
                if differCount == 1:
                    assert len(matched) == (len(variableBefore['value']) - 1)
                    for i in range(len(variableAfter)):
                        if variableAfter['value'][i] not in matched:
                            assert variableAfter['value'][i] not in variableBefore['value']
                            # assert variableAfter['value'][i]['key'] == candidateKey and variableAfter['value'][i]['value'] != candidateValue
                            if (variableAfter['value'][i]['key'] == candidateKey) and (variableAfter['value'][i]['value'] != candidateValue):
                                res.append(candidateKey)
                                return (res + getDifferentEntryTrace(candidateValue, variableAfter['value'][i]['value']))
                            break
        return res
    
    def entryChangeTo(lst, variableAfter):
        cur = copy.deepcopy(variableAfter)
        for e in lst:
            if isinstance(e, int): # list
                assert cur['type'] == 'list'
                cur = cur['value'][e]
            else: # dict
                assert cur['type'] == 'dict'
                for kv in cur['value']:
                    if kv['key'] == e:
                        cur = kv['value']
                        break
        return verbose_string(cur)

    def about(ctx):
        nxt = ctx["next"]
        if nxt["type"] == "Frame":
            return f"about to run method {nxt['name']} with argument {verbose_string(nxt['value'])}"
        elif nxt["type"] == "Load":
            return f"about to load variable {nxt['var']}"
        elif nxt["type"] == "Store":
            return f"about to store: {nxt['var']}<-{verbose_string(nxt['value'])}"
        elif nxt["type"] == "Print":
            return f"about to print {verbose_string(nxt['value'])}"
        elif nxt["type"] == "AtomicInc":
            return "about to execute atomic section"
        elif nxt["type"] == "Assert":
            return "assertion failed"
        else:
            return f"about to {nxt['type']}"
        
    # with open(filenames['hco'], 'r') as hcoFile:
    #     hco = json.load(hcoFile, strict=False)
    hvm = hco['hvm']

    # construct summaryMain.microSteps to be a list of microsteps
    constructMicrosteps()
    # construct summaryMain.threadMode
    constructThreadMode()
    # construct stackTraceText to be stack trace to display at each microstep
    constructStackTraceTextList()
    # construct stackTopDisplay
    constructStackTopDisplay()

    return friendlyOutput()



def bag_add(bag, item):
    cnt = bag.get(item)
    if cnt is None:
        bag[item] = 1
    else:
        bag[item] = cnt + 1

def bag_remove(bag, item):
    cnt = bag[item]
    assert cnt > 0
    if cnt == 1:
        del bag[item]
    else:
        bag[item] = cnt - 1


from typing import Any, List, NamedTuple

class ErrorToken(NamedTuple):
    line: int
    message: str
    column: int
    lexeme: str
    filename: str
    is_eof_error: bool

class HarmonyCompilerErrorCollection(Exception):
    def __init__(self, errors: List[ErrorToken]) -> None:
        super().__init__()
        self.errors = errors

class HarmonyCompilerError(Exception):
    """
    Error encountered during the compilation of a Harmony program.
    """

    def __init__(self, message: str, filename: str = None, line: int = None,
                 column: int = None, lexeme: Any = None, is_eof_error=False):
        super().__init__()
        self.message = message
        self.token = ErrorToken(
            message=message,
            filename=filename or "",
            line=line or 0,
            column=column or 0,
            lexeme=str(lexeme),
            is_eof_error=is_eof_error
        )
